Package: causalsim
Title: Simulation-Ready Causal Data Generating Processes
Version: 0.1.0
Authors@R: 
    person("Chayce", "Reed", email = "Chayce.Reed.HSE@dartmouth.edu", role = c("aut", "cre"))
Description: Construct, simulate, and evaluate causal data generating processes
    (DGPs) with known ground truth. Designed for benchmarking causal estimators,
    studying confounding and treatment-effect heterogeneity, and building
    reproducible teaching examples. Covariate roles (confounder, effect modifier,
    noise) and heterogeneous treatment effects are first-class concepts in the
    API, and estimator performance is summarised with bias, root mean squared
    error, confidence-interval coverage, and power.
License: MIT + file LICENSE
URL: https://chaycereed.github.io/causalsim/,
        https://github.com/chaycereed/causalsim
BugReports: https://github.com/chaycereed/causalsim/issues
Encoding: UTF-8
Depends: R (>= 4.0.0)
RoxygenNote: 7.3.3
Suggests: knitr, pkgdown, rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/Needs/website: pkgdown
NeedsCompilation: no
Packaged: 2026-08-21 12:34:36 UTC; chayce
Author: Chayce Reed [aut, cre]
Maintainer: Chayce Reed <Chayce.Reed.HSE@dartmouth.edu>
Repository: CRAN
Date/Publication: 2026-08-30 10:00:15 UTC

# Generated by roxygen2: do not edit by hand

S3method(plot,causalsim_eval)
S3method(print,causalsim_covar)
S3method(print,causalsim_dgp)
S3method(print,causalsim_eval)
S3method(print,causalsim_eval_grid)
S3method(print,causalsim_eval_summary)
S3method(summary,causalsim_eval)
export(causalsim)
export(causalsim_covar)
export(causalsim_dgp)
export(causalsim_draw)
export(causalsim_eval)
export(causalsim_eval_grid)

# causalsim

An R package for defining causal data generating processes with known ground truth and evaluating estimator performance against them.

## Features

- Structural causal model with explicit effect, propensity, and baseline functions
- Named covariate roles: confounder, effect modifier, noise
- Preset confounding levels (`"low"`, `"moderate"`, `"high"`) or custom functions
- Exact or Monte Carlo true ATE computed at construction time
- Flexible estimator interface: named numeric vector, named list, or one-row data frame
- Tidy performance metrics: bias, RMSE, coverage, and power with Monte Carlo standard errors
- Grid evaluation over the Cartesian product of any DGP parameters
- Reproducible via seed control at every stage

## Installation

Requires R 4.0 or higher. Install from GitHub:

```r
# install.packages("devtools")
devtools::install_github("chaycereed/causalsim")
```

## Usage

### Quick start

`causalsim()` generates a dataset in one call. The returned data frame includes covariate columns, treatment `A`, outcome `Y`, individual effect `.tau`, and true propensity `.p`.

```r
library(causalsim)

data <- causalsim(n = 500, n_confounders = 1, effect = 2, seed = 1L)
head(data)
```

For estimator benchmarking or multi-draw workflows, use `causalsim_dgp()` and `causalsim_draw()` directly (see below).

### Define a data generating process

`causalsim_dgp()` specifies the structural model. Parameters can be scalars, preset strings, or functions of the covariate names.

```r
dgp <- causalsim_dgp(
  n = 500,
  n_confounders = 1,
  effect = 2,
  propensity = "moderate",
  baseline = "moderate"
)
dgp
```

### Draw a dataset

`causalsim_draw()` simulates one dataset from the DGP. The returned data frame includes covariate columns, treatment `A`, outcome `Y`, individual effect `.tau`, and propensity `.p`.

```r
dat <- causalsim_draw(dgp, seed = 1L)
head(dat)
```

### Evaluate an estimator

An estimator is any function that accepts a data frame and returns a named numeric vector with at minimum an `estimate` field. `ci_lower` and `ci_upper` enable coverage and power metrics.

```r
ols_est <- function(data) {
  fit <- lm(Y ~ A + W, data = data)
  est <- coef(fit)[["A"]]
  se <- sqrt(vcov(fit)["A", "A"])
  c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
}

result <- causalsim_eval(dgp, ols_est, reps = 200L, seed = 1L)
result

summary(result)
plot(result)
```

### Evaluate across a parameter grid

`causalsim_eval_grid()` runs the evaluator over the Cartesian product of any DGP parameters, returning a tidy data frame of metrics for each cell.

```r
grid_result <- causalsim_eval_grid(
  dgp = dgp,
  estimator = ols_est,
  vary = list(n = c(100L, 250L, 500L, 1000L)),
  reps = 200L,
  metrics = c("bias", "rmse"),
  seed = 1L
)
grid_result
```

### Heterogeneous effects

Declare a covariate with `role = "effect_modifier"` and reference it in a
function passed to `effect` to make the treatment effect vary across subgroups.
The individual effects are stored in the `.tau` column as ground truth.

```r
het_dgp <- causalsim_dgp(
  n = 2000,
  covariates = list(
    W = causalsim_covar("normal", role = "confounder"),
    V = causalsim_covar("binary", role = "effect_modifier", prob = 0.5)
  ),
  effect     = function(V) 2 + 3 * V   # effect is 2 when V = 0, 5 when V = 1
)

d <- causalsim_draw(het_dgp, seed = 1L)
tapply(d$.tau, d$V, mean)              # 2 for V = 0, 5 for V = 1
```

The function passed to `effect` is what activates the modifier; a covariate
labelled `effect_modifier` that `effect` never references is inert, and
`causalsim_dgp()` warns when that happens.

## Roadmap

Planned for a future release:

- **Additional covariate roles.** The current roles (`confounder`, `effect_modifier`, `noise`) cover the common cases. Two single-path roles would complete the treatment/outcome taxonomy:
  - `instrument` — drives treatment only (enters the propensity model, excluded from the outcome), for benchmarking instrumental-variable estimators.
  - `prognostic` — drives the outcome only (enters the baseline, independent of treatment), for studying precision covariates and variance reduction.

  These will ship together with worked examples that demonstrate each (an IV estimator and a variance-reduction comparison, respectively).
- **Assumption-violation helpers.** First-class support for unmeasured confounding and positivity violations.

## License

MIT License. See `LICENSE` for details.

#' Evaluate a causal estimator against a known DGP
#'
#' Repeatedly draws datasets from a [causalsim_dgp()] object, applies a
#' user-supplied estimator to each draw, and returns a tidy summary of
#' estimator performance against the known ground truth.
#'
#' @param dgp A `causalsim_dgp` object.
#' @param estimator A function that accepts a data frame produced by
#'   [causalsim_draw()] and returns a named numeric vector or single-row
#'   data frame. Must include an `estimate` field. For `"coverage"` and
#'   `"power"` metrics, must also include `ci_lower` and `ci_upper`. See
#'   Details.
#' @param reps Positive integer. Number of simulation replications.
#'   Default `200L`.
#' @param metrics Character vector. Subset of `"bias"`, `"rmse"`,
#'   `"coverage"`, `"power"`. Default: all four.
#' @param seed Integer or `NULL`. Passed to [set.seed()] before the first
#'   replication. Default `NULL`.
#'
#' @details
#' ## Estimator convention
#'
#' The estimator receives the full data frame returned by [causalsim_draw()],
#' including ground-truth columns `.tau` and `.p`. It should return a named
#' numeric vector or one-row data frame with at minimum:
#'
#' ```r
#' my_estimator <- function(data) {
#'   fit <- lm(Y ~ A + W, data = data)
#'   est <- coef(fit)["A"]
#'   se <- sqrt(vcov(fit)["A", "A"])
#'   c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
#' }
#' ```
#'
#' ## Metrics
#'
#' \describe{
#'   \item{`"bias"`}{Monte Carlo estimate of bias: `mean(estimate) - true_ate`.}
#'   \item{`"rmse"`}{Root mean squared error: `sqrt(mean((estimate - true_ate)^2))`.}
#'   \item{`"coverage"`}{Proportion of replications where the CI brackets
#'     `true_ate`. Requires `ci_lower` and `ci_upper`.}
#'   \item{`"power"`}{Proportion of replications where the CI excludes zero,
#'     i.e., the rejection rate of H0: ATE = 0. When `true_ate != 0`,
#'     this is power; when `true_ate == 0`, it is the Type I error rate.
#'     Requires `ci_lower` and `ci_upper`.}
#' }
#'
#' Monte Carlo standard errors (MCSE) are reported alongside each metric.
#' For bias: `sd(estimates) / sqrt(reps)`. For RMSE: delta-method
#' approximation. For coverage and power: Bernoulli SE.
#'
#' ## Raw draws
#'
#' Per-replication estimates are stored in `result$draws` and can be used
#' for custom analysis or plotting.
#'
#' @return An S3 object of class `causalsim_eval` with components:
#' \describe{
#'   \item{`metrics`}{Tidy data frame: `metric`, `value`, `se`.}
#'   \item{`draws`}{Data frame of per-replication estimates, one row per rep.}
#'   \item{`true_ate`}{The DGP's true ATE.}
#'   \item{`reps`}{Number of replications run.}
#' }
#'
#' @examples
#' dgp <- causalsim_dgp(n = 300, n_confounders = 1, effect = 2,
#'                      propensity = 0.5)
#'
#' # OLS estimator (unbiased under RCT)
#' ols_estimator <- function(data) {
#'   fit <- lm(Y ~ A + W, data = data)
#'   est <- coef(fit)[["A"]]
#'   se <- sqrt(vcov(fit)["A", "A"])
#'   c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
#' }
#'
#' \donttest{
#' result <- causalsim_eval(dgp, ols_estimator, reps = 100L, seed = 1L)
#' result
#' }
#'
#' @export
causalsim_eval <- function(
  dgp,
  estimator,
  reps = 200L,
  metrics = c("bias", "rmse", "coverage", "power"),
  seed = NULL
) {
  if (!inherits(dgp, "causalsim_dgp")) {
    stop("`dgp` must be a causalsim_dgp object.", call. = FALSE)
  }
  if (!is.function(estimator)) {
    stop("`estimator` must be a function.", call. = FALSE)
  }
  reps <- .validate_positive(reps, "reps", as_int = TRUE)
  metrics <- match.arg(
    metrics,
    choices = c("bias", "rmse", "coverage", "power"),
    several.ok = TRUE
  )

  if (!is.null(seed)) set.seed(seed)

  draws <- .collect_draws(dgp, estimator, reps)

  needs_ci <- any(metrics %in% c("coverage", "power"))
  if (needs_ci && !all(c("ci_lower", "ci_upper") %in% names(draws))) {
    stop(
      paste0(
        "Metrics 'coverage' and 'power' require the estimator to return ",
        "'ci_lower' and 'ci_upper'."
      ),
      call. = FALSE
    )
  }

  metrics_df <- .compute_metrics(draws, dgp$true_ate, metrics)

  structure(
    list(
      metrics = metrics_df,
      draws = draws,
      true_ate = dgp$true_ate,
      reps = reps
    ),
    class = "causalsim_eval"
  )
}

#' @export
print.causalsim_eval <- function(x, digits = 4L, ...) {
  cat(sprintf(
    "<causalsim_eval>  reps: %d  true ATE: %s\n\n",
    x$reps, format(round(x$true_ate, digits))
  ))
  m <- x$metrics
  m$value <- round(m$value, digits)
  m$se <- round(m$se, digits + 1L)
  print(m, row.names = FALSE)
  invisible(x)
}

#' Summarise a causalsim_eval result
#'
#' Returns a structured summary of estimator performance including the metrics
#' table and the distribution of per-replication estimates (mean, SD, median,
#' and 10th / 90th percentiles).
#'
#' @param object A `causalsim_eval` object.
#' @param ... Ignored.
#'
#' @return An S3 object of class `causalsim_eval_summary` with components:
#' \describe{
#'   \item{`metrics`}{The metrics data frame from the original result.}
#'   \item{`mean_estimate`}{Mean of per-replication estimates.}
#'   \item{`sd_estimate`}{Standard deviation of per-replication estimates.}
#'   \item{`median_estimate`}{Median of per-replication estimates.}
#'   \item{`p10_estimate`}{10th percentile of per-replication estimates.}
#'   \item{`p90_estimate`}{90th percentile of per-replication estimates.}
#'   \item{`true_ate`}{True ATE from the DGP.}
#'   \item{`reps`}{Number of replications.}
#' }
#'
#' @export
summary.causalsim_eval <- function(object, ...) {
  est <- object$draws$estimate
  structure(
    list(
      metrics = object$metrics,
      mean_estimate = mean(est),
      sd_estimate = stats::sd(est),
      median_estimate = stats::median(est),
      p10_estimate = unname(stats::quantile(est, 0.10)),
      p90_estimate = unname(stats::quantile(est, 0.90)),
      true_ate = object$true_ate,
      reps = object$reps
    ),
    class = "causalsim_eval_summary"
  )
}

#' @export
print.causalsim_eval_summary <- function(x, digits = 4L, ...) {
  cat(sprintf(
    "<causalsim_eval_summary>  reps: %d  true ATE: %s\n\n",
    x$reps, format(round(x$true_ate, digits))
  ))
  cat("Estimate distribution:\n")
  cat(sprintf("  mean  : %s\n", format(round(x$mean_estimate, digits))))
  cat(sprintf("  sd    : %s\n", format(round(x$sd_estimate, digits))))
  cat(sprintf("  median: %s\n", format(round(x$median_estimate, digits))))
  cat(sprintf("  [p10, p90]: [%s, %s]\n\n",
              format(round(x$p10_estimate, digits)),
              format(round(x$p90_estimate, digits))))
  cat("Metrics:\n")
  m <- x$metrics
  m$value <- round(m$value, digits)
  m$se <- round(m$se, digits + 1L)
  print(m, row.names = FALSE)
  invisible(x)
}

#' Plot the distribution of estimates from a causalsim_eval result
#'
#' Draws a histogram of per-replication estimates with vertical lines marking
#' the true ATE (solid) and the mean estimate (dashed).
#'
#' @param x A `causalsim_eval` object.
#' @param ... Additional arguments passed to [graphics::hist()].
#'
#' @return `x`, invisibly.
#'
#' @export
plot.causalsim_eval <- function(x, ...) {
  est <- x$draws$estimate
  true_ate <- x$true_ate
  mean_est <- mean(est)

  graphics::hist(
    est,
    main = "Distribution of Estimates",
    xlab = "Estimate",
    col = "grey85",
    border = "white",
    ...
  )
  graphics::abline(v = true_ate, lwd = 2)
  graphics::abline(v = mean_est, lwd = 1, lty = 2)
  graphics::legend(
    "topright",
    legend = c(
      sprintf("true ATE = %.3f", true_ate),
      sprintf("mean est = %.3f", mean_est)
    ),
    lty = c(1L, 2L),
    lwd = c(2L, 1L),
    bty = "n"
  )
  invisible(x)
}

# ── Shared fixtures ───────────────────────────────────────────────────────────

# DGP: RCT (propensity = 0.5) so OLS is unbiased; true ATE = 2
dgp_rct <- causalsim_dgp(
  n             = 300,
  n_confounders = 1,
  effect        = 2,
  propensity    = 0.5
)

# Constant estimator: always returns exactly the true ATE with a wide CI
const_estimator <- function(data) {
  c(estimate = 2, ci_lower = 1, ci_upper = 3)
}

# Biased estimator: always overshoots by 0.5
biased_estimator <- function(data) {
  c(estimate = 2.5, ci_lower = 2, ci_upper = 3)
}

# Estimator returning a data frame instead of a vector
df_estimator <- function(data) {
  data.frame(estimate = 2, ci_lower = 1, ci_upper = 3)
}

# Estimator with no CI (bias/rmse only)
noci_estimator <- function(data) {
  c(estimate = 2)
}

# ── Return type and structure ─────────────────────────────────────────────────

test_that("causalsim_eval returns causalsim_eval class", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 10L,
                           seed = 1L)
  expect_s3_class(result, "causalsim_eval")
})

test_that("result$metrics has metric, value, se columns", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 10L, seed = 1L)
  expect_named(result$metrics, c("metric", "value", "se"))
})

test_that("result$draws has one row per rep", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 20L, seed = 1L)
  expect_equal(nrow(result$draws), 20L)
})

test_that("result$draws has estimate column", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 10L, seed = 1L)
  expect_true("estimate" %in% names(result$draws))
})

test_that("result$reps matches requested reps", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 15L, seed = 1L)
  expect_equal(result$reps, 15L)
})

test_that("result$true_ate matches dgp$true_ate", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 10L, seed = 1L)
  expect_equal(result$true_ate, dgp_rct$true_ate)
})

# ── Metric values ─────────────────────────────────────────────────────────────

test_that("bias is 0 for constant-exact estimator", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 50L,
                           metrics = "bias", seed = 1L)
  expect_equal(result$metrics$value[result$metrics$metric == "bias"], 0)
})

test_that("bias is approximately 0.5 for consistently biased estimator", {
  result <- causalsim_eval(dgp_rct, biased_estimator, reps = 50L,
                           metrics = "bias", seed = 1L)
  expect_equal(
    result$metrics$value[result$metrics$metric == "bias"], 0.5,
    tolerance = 1e-10
  )
})

test_that("rmse is 0 for constant-exact estimator", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 50L,
                           metrics = "rmse", seed = 1L)
  expect_equal(result$metrics$value[result$metrics$metric == "rmse"], 0)
})

test_that("coverage is 1 when CI always brackets true_ate", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 50L,
                           metrics = "coverage", seed = 1L)
  expect_equal(
    result$metrics$value[result$metrics$metric == "coverage"], 1
  )
})

test_that("power is 1 when CI always excludes zero", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 50L,
                           metrics = "power", seed = 1L)
  expect_equal(
    result$metrics$value[result$metrics$metric == "power"], 1
  )
})

test_that("only requested metrics are returned", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 10L,
                           metrics = c("bias", "rmse"), seed = 1L)
  expect_equal(sort(result$metrics$metric), c("bias", "rmse"))
})

# ── Data frame estimator ──────────────────────────────────────────────────────

test_that("estimator returning a data frame is accepted", {
  expect_no_error(
    causalsim_eval(dgp_rct, df_estimator, reps = 10L,
                   metrics = "bias", seed = 1L)
  )
})

# ── Named list estimator ──────────────────────────────────────────────────────

test_that("estimator returning a named numeric list is accepted", {
  list_estimator <- function(data) {
    list(estimate = 2, ci_lower = 1, ci_upper = 3)
  }
  expect_no_error(
    causalsim_eval(dgp_rct, list_estimator, reps = 10L,
                   metrics = "bias", seed = 1L)
  )
})

test_that("list estimator gives same result as equivalent vector estimator", {
  vec_est  <- function(data) c(estimate = 2, ci_lower = 1, ci_upper = 3)
  list_est <- function(data) list(estimate = 2, ci_lower = 1, ci_upper = 3)
  r1 <- causalsim_eval(dgp_rct, vec_est,  reps = 10L,
                       metrics = "bias", seed = 5L)
  r2 <- causalsim_eval(dgp_rct, list_est, reps = 10L,
                       metrics = "bias", seed = 5L)
  expect_equal(r1$metrics, r2$metrics)
})

test_that("inner names on list values do not override outer names", {
  inner_named_est <- function(data) {
    fit <- lm(Y ~ A, data = data)
    list(
      estimate = coef(fit)["A"],
      ci_lower = confint(fit)["A", 1],
      ci_upper = confint(fit)["A", 2]
    )
  }
  result <- causalsim_eval(dgp_rct, inner_named_est, reps = 5L,
                           metrics = "bias", seed = 1L)
  expect_true("bias" %in% result$metrics$metric)
})

test_that("list estimator with non-numeric element errors", {
  bad_list_est <- function(data) {
    list(estimate = 2, ci_lower = "1", ci_upper = 3)
  }
  expect_error(
    causalsim_eval(dgp_rct, bad_list_est, reps = 5L, seed = 1L),
    "not all elements are named numeric"
  )
})

# ── No-CI estimator ───────────────────────────────────────────────────────────

test_that("bias and rmse work with no-CI estimator", {
  expect_no_error(
    causalsim_eval(dgp_rct, noci_estimator, reps = 10L,
                   metrics = c("bias", "rmse"), seed = 1L)
  )
})

test_that("requesting coverage with no-CI estimator errors", {
  expect_error(
    causalsim_eval(dgp_rct, noci_estimator, reps = 10L,
                   metrics = "coverage", seed = 1L),
    "ci_lower"
  )
})

# ── Reproducibility ───────────────────────────────────────────────────────────

test_that("same seed gives identical results", {
  r1 <- causalsim_eval(dgp_rct, const_estimator, reps = 20L, seed = 7L)
  r2 <- causalsim_eval(dgp_rct, const_estimator, reps = 20L, seed = 7L)
  expect_identical(r1$draws, r2$draws)
})

# ── Input validation ──────────────────────────────────────────────────────────

test_that("non-causalsim_dgp input errors", {
  expect_error(causalsim_eval(list(), const_estimator), "causalsim_dgp")
})

test_that("non-function estimator errors", {
  expect_error(causalsim_eval(dgp_rct, "lm"), "`estimator`")
})

test_that("estimator missing 'estimate' field errors", {
  bad_estimator <- function(data) c(coef = 2, lower = 1, upper = 3)
  expect_error(
    causalsim_eval(dgp_rct, bad_estimator, reps = 5L, seed = 1L),
    "'estimate'"
  )
})

test_that("invalid metric name errors", {
  expect_error(
    causalsim_eval(dgp_rct, const_estimator, reps = 5L,
                   metrics = "variance"),
    "should be one of"
  )
})

# ── Print method ──────────────────────────────────────────────────────────────

test_that("print method runs without error and returns invisibly", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 10L, seed = 1L)
  expect_invisible(print(result))
  expect_output(print(result), "causalsim_eval")
  expect_output(print(result), "true ATE")
})

# ── Summary method ────────────────────────────────────────────────────────────

test_that("summary returns causalsim_eval_summary class", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 20L, seed = 1L)
  expect_s3_class(summary(result), "causalsim_eval_summary")
})

test_that("summary has expected fields", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 20L, seed = 1L)
  s <- summary(result)
  expect_true(all(c("mean_estimate", "sd_estimate", "median_estimate",
                    "p10_estimate", "p90_estimate",
                    "true_ate", "reps", "metrics") %in% names(s)))
})

test_that("summary mean_estimate is correct for constant estimator", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 20L, seed = 1L)
  s <- summary(result)
  expect_equal(s$mean_estimate, 2)
})

test_that("summary prints without error and returns invisibly", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 10L, seed = 1L)
  s <- summary(result)
  expect_invisible(print(s))
  expect_output(print(s), "mean")
  expect_output(print(s), "true ATE")
})

# ── Plot method ───────────────────────────────────────────────────────────────

test_that("plot runs without error and returns invisibly", {
  result <- causalsim_eval(dgp_rct, const_estimator, reps = 20L, seed = 1L)
  grDevices::pdf(NULL)
  on.exit(grDevices::dev.off())
  expect_invisible(plot(result))
})

#' Evaluate an estimator across a parameter grid
#'
#' Runs [causalsim_eval()] over the Cartesian product of one or more DGP
#' parameter values, returning a tidy data frame of performance metrics for
#' every cell. Designed for studying how estimator behavior changes with
#' sample size, confounding strength, effect size, or noise level.
#'
#' @param dgp A `causalsim_dgp` object. Provides fixed parameter values for
#'   all dimensions not listed in `vary`.
#' @param estimator A function as accepted by [causalsim_eval()].
#' @param vary Named list of atomic vectors. Each name must be a valid
#'   [causalsim_dgp()] argument (except `covariates`, which cannot be
#'   varied atomically). Each vector supplies the values to try for that
#'   dimension. The full grid is the Cartesian product of all dimensions.
#' @param reps Positive integer. Replications per grid cell. Default `200L`.
#' @param metrics Character vector. Passed to [causalsim_eval()].
#' @param seed Integer or `NULL`. Passed to [set.seed()] before the first
#'   cell. The same RNG stream continues across cells, so results are
#'   jointly reproducible. Default `NULL`.
#' @param verbose Logical. If `TRUE`, prints a progress message before each
#'   cell. Default `FALSE`.
#'
#' @details
#' ## How it works
#'
#' For each cell in `expand.grid(vary)`, `causalsim_eval_grid()` takes `dgp`'s
#' stored original parameters, overrides the cell's values, reconstructs a
#' new [causalsim_dgp()], runs [causalsim_eval()], and tags the resulting
#' metrics with the cell's parameter values.
#'
#' ## Constraints on `vary`
#'
#' Elements of `vary` must be atomic vectors (character, numeric, integer).
#' Functions cannot be varied via this interface. `covariates` (a list of
#' [causalsim_covar()] objects) is excluded. For complex covariate variation,
#' construct DGPs manually and use [causalsim_eval()] directly.
#'
#' Note that varying `n_confounders` from 1 to 2 changes auto-generated
#' covariate names from `W` to `W1, W2`. If the base DGP's `effect` function
#' references `W`, the reconstructed DGP will error at construction time.
#' Design the base DGP accordingly.
#'
#' @return An S3 object of class `causalsim_eval_grid` with components:
#' \describe{
#'   \item{`results`}{Tidy data frame: one row per (cell, metric).
#'     Columns: grid parameter names, `metric`, `value`, `se`.}
#'   \item{`grid`}{Data frame of grid cells, one row per cell.}
#'   \item{`vary`}{Character vector of varied parameter names.}
#'   \item{`reps`}{Replications per cell.}
#'   \item{`metrics`}{Metrics evaluated.}
#' }
#'
#' @examples
#' dgp <- causalsim_dgp(n = 500, n_confounders = 1, effect = 2,
#'                      propensity = 0.5)
#'
#' ols_estimator <- function(data) {
#'   fit <- lm(Y ~ A + W, data = data)
#'   est <- coef(fit)[["A"]]
#'   se  <- sqrt(vcov(fit)["A", "A"])
#'   c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
#' }
#'
#' \donttest{
#' grid_result <- causalsim_eval_grid(
#'   dgp = dgp,
#'   estimator = ols_estimator,
#'   vary = list(n = c(250L, 500L, 1000L)),
#'   reps = 50L,
#'   seed = 1L
#' )
#' grid_result
#' }
#'
#' @export
causalsim_eval_grid <- function(
  dgp,
  estimator,
  vary,
  reps = 200L,
  metrics = c("bias", "rmse", "coverage", "power"),
  seed = NULL,
  verbose = FALSE
) {
  if (!inherits(dgp, "causalsim_dgp")) {
    stop("`dgp` must be a causalsim_dgp object.", call. = FALSE)
  }
  if (!is.function(estimator)) {
    stop("`estimator` must be a function.", call. = FALSE)
  }
  reps <- .validate_positive(reps, "reps", as_int = TRUE)

  .validate_vary(vary, dgp)

  grid_df <- expand.grid(vary, stringsAsFactors = FALSE)
  n_cells <- nrow(grid_df)
  vary_nms <- names(vary)

  metrics <- match.arg(
    metrics,
    choices = c("bias", "rmse", "coverage", "power"),
    several.ok = TRUE
  )

  if (!is.null(seed)) set.seed(seed)

  all_rows <- vector("list", n_cells)
  for (i in seq_len(n_cells)) {
    if (verbose) {
      message(sprintf("Grid cell %d / %d", i, n_cells))
    }

    cell_params <- dgp$params
    for (nm in vary_nms) {
      cell_params[[nm]] <- grid_df[[nm]][[i]]
    }

    cell_dgp <- do.call(causalsim_dgp, cell_params)
    cell_eval <- causalsim_eval(cell_dgp, estimator,
                                  reps = reps, metrics = metrics)
    cell_result <- cell_eval$metrics

    for (nm in vary_nms) {
      cell_result[[nm]] <- grid_df[[nm]][[i]]
    }

    all_rows[[i]] <- cell_result
  }

  results <- do.call(rbind, all_rows)
  results <- results[, c(vary_nms, "metric", "value", "se")]
  rownames(results) <- NULL

  structure(
    list(
      results = results,
      grid = grid_df,
      vary = vary_nms,
      reps = reps,
      metrics = metrics
    ),
    class = "causalsim_eval_grid"
  )
}

#' @export
print.causalsim_eval_grid <- function(x, n = 10L, ...) {
  n_cells <- nrow(x$grid)
  cat(sprintf(
    "<causalsim_eval_grid>  %d cell%s  vary: %s  reps/cell: %d\n",
    n_cells,
    if (n_cells == 1L) "" else "s",
    paste(x$vary, collapse = ", "),
    x$reps
  ))
  cat(sprintf("  metrics: %s\n\n", paste(x$metrics, collapse = ", ")))
  nr <- nrow(x$results)
  print(x$results[seq_len(min(n, nr)), ], row.names = FALSE)
  if (nr > n) {
    cat(sprintf(
      "\n  ... %d more row%s. Access full results via $results.\n",
      nr - n, if (nr - n == 1L) "" else "s"
    ))
  }
  invisible(x)
}

# ── Internal ──────────────────────────────────────────────────────────────────

.validate_vary <- function(vary, dgp) {
  if (!is.list(vary) || length(vary) == 0L) {
    stop(
      "`vary` must be a non-empty named list of atomic vectors.",
      call. = FALSE
    )
  }
  if (is.null(names(vary)) || any(names(vary) == "")) {
    stop("Every element of `vary` must be named.", call. = FALSE)
  }

  valid_nms <- setdiff(names(formals(causalsim_dgp)), "covariates")
  invalid <- setdiff(names(vary), valid_nms)
  if (length(invalid) > 0L) {
    stop(
      paste0(
        "`vary` name(s) not a valid causalsim_dgp() argument: ",
        paste(invalid, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  not_atomic <- !vapply(vary, is.atomic, logical(1L))
  if (any(not_atomic)) {
    stop(
      paste0(
        "All elements of `vary` must be atomic vectors. Non-atomic: ",
        paste(names(vary)[not_atomic], collapse = ", ")
      ),
      call. = FALSE
    )
  }
}

# ── Shared fixtures ───────────────────────────────────────────────────────────

dgp_base <- causalsim_dgp(
  n             = 100,
  n_confounders = 1,
  effect        = 2,
  propensity    = 0.5
)

# Constant estimator: deterministic, CI always brackets true ATE = 2
const_est <- function(data) {
  c(estimate = 2, ci_lower = 1, ci_upper = 3)
}

# ── Return type and structure ─────────────────────────────────────────────────

test_that("causalsim_eval_grid returns correct S3 class", {
  g <- causalsim_eval_grid(dgp_base, const_est,
                      vary = list(n = c(100L, 200L)),
                      reps = 5L, metrics = "bias", seed = 1L)
  expect_s3_class(g, "causalsim_eval_grid")
})

test_that("results has vary columns + metric + value + se", {
  g <- causalsim_eval_grid(dgp_base, const_est,
                      vary = list(n = c(100L, 200L)),
                      reps = 5L, metrics = "bias", seed = 1L)
  expect_true(all(c("n", "metric", "value", "se") %in% names(g$results)))
})

test_that("results has n_cells x n_metrics rows", {
  g <- causalsim_eval_grid(
    dgp_base, const_est,
    vary    = list(n = c(100L, 200L), sigma = c(0.5, 1.0)),
    reps    = 5L,
    metrics = c("bias", "rmse"),
    seed    = 1L
  )
  expect_equal(nrow(g$results), 8L)  # 2 x 2 cells x 2 metrics
})

test_that("grid parameter values appear correctly in results", {
  g <- causalsim_eval_grid(dgp_base, const_est,
                      vary = list(n = c(100L, 200L, 300L)),
                      reps = 5L, metrics = "bias", seed = 1L)
  expect_setequal(unique(g$results$n), c(100L, 200L, 300L))
})

test_that("vary parameter names appear in $vary", {
  g <- causalsim_eval_grid(
    dgp_base, const_est,
    vary = list(n = c(100L, 200L), sigma = c(0.5, 1.0)),
    reps = 5L, metrics = "bias", seed = 1L
  )
  expect_equal(g$vary, c("n", "sigma"))
})

test_that("$grid has one row per cell", {
  g <- causalsim_eval_grid(
    dgp_base, const_est,
    vary = list(n = c(100L, 200L), sigma = c(0.5, 1.0)),
    reps = 5L, metrics = "bias", seed = 1L
  )
  expect_equal(nrow(g$grid), 4L)
})

test_that("reps is stored on the result", {
  g <- causalsim_eval_grid(dgp_base, const_est,
                      vary = list(n = c(100L, 200L)),
                      reps = 7L, metrics = "bias", seed = 1L)
  expect_equal(g$reps, 7L)
})

# ── Metric values ─────────────────────────────────────────────────────────────

test_that("bias is 0 for constant-exact estimator across all cells", {
  g <- causalsim_eval_grid(dgp_base, const_est,
                      vary = list(n = c(100L, 200L)),
                      reps = 5L, metrics = "bias", seed = 1L)
  bias_vals <- g$results$value[g$results$metric == "bias"]
  expect_true(all(bias_vals == 0))
})

# ── Reproducibility ───────────────────────────────────────────────────────────

test_that("same seed gives identical results", {
  g1 <- causalsim_eval_grid(dgp_base, const_est,
                       vary = list(n = c(100L, 200L)),
                       reps = 5L, metrics = "bias", seed = 42L)
  g2 <- causalsim_eval_grid(dgp_base, const_est,
                       vary = list(n = c(100L, 200L)),
                       reps = 5L, metrics = "bias", seed = 42L)
  expect_identical(g1$results, g2$results)
})

# ── Input validation ──────────────────────────────────────────────────────────

test_that("non-causalsim_dgp input errors", {
  expect_error(
    causalsim_eval_grid(list(), const_est, vary = list(n = 100L), reps = 5L),
    "causalsim_dgp"
  )
})

test_that("non-function estimator errors", {
  expect_error(
    causalsim_eval_grid(dgp_base, "ols", vary = list(n = 100L), reps = 5L),
    "`estimator`"
  )
})

test_that("empty vary errors", {
  expect_error(
    causalsim_eval_grid(dgp_base, const_est, vary = list(), reps = 5L),
    "non-empty"
  )
})

test_that("unnamed vary errors", {
  expect_error(
    causalsim_eval_grid(dgp_base, const_est,
                   vary = list(c(100L, 200L)), reps = 5L),
    "named"
  )
})

test_that("invalid vary parameter name errors", {
  expect_error(
    causalsim_eval_grid(dgp_base, const_est,
                   vary = list(foo = c(1, 2)), reps = 5L),
    "not a valid"
  )
})

test_that("non-atomic vary element errors", {
  expect_error(
    causalsim_eval_grid(
      dgp_base, const_est,
      vary = list(effect = list(function(W) W, function(W) 2 * W)),
      reps = 5L
    ),
    "atomic"
  )
})

# ── Print method ──────────────────────────────────────────────────────────────

test_that("print runs without error and returns invisibly", {
  g <- causalsim_eval_grid(dgp_base, const_est,
                      vary = list(n = c(100L, 200L)),
                      reps = 5L, metrics = "bias", seed = 1L)
  expect_invisible(print(g))
  expect_output(print(g), "causalsim_eval_grid")
  expect_output(print(g), "vary")
})

#' Define a covariate for a causal DGP
#'
#' Constructs a fully-specified covariate descriptor for use in
#' [causalsim_dgp()]. Every covariate has a distribution family, optional
#' distribution parameters, and one or more causal roles that control how it
#' enters the automatically-generated treatment and outcome models.
#'
#' @param dist Character. Distribution family. One of `"normal"`, `"binary"`,
#'   `"uniform"`.
#' @param role Character vector. Causal role(s). One or more of:
#'   \describe{
#'     \item{`"confounder"`}{Enters both the propensity model and the outcome
#'       baseline; creates confounding bias in naive estimators.}
#'     \item{`"effect_modifier"`}{Available for use in the `effect` function;
#'       excluded from the propensity model unless also a `"confounder"`.}
#'     \item{`"noise"`}{Independent of both treatment and outcome; adds
#'       variance without confounding.}
#'   }
#'   Roles are not mutually exclusive: `role = c("confounder", "effect_modifier")`
#'   specifies a variable that both confounds and moderates the effect.
#' @param ... Distribution parameters.
#'   - `"normal"`: `mean` (default 0), `sd` (default 1)
#'   - `"binary"`: `prob` (default 0.5)
#'   - `"uniform"`: `min` (default 0), `max` (default 1)
#'
#' @return An S3 object of class `causalsim_covar`.
#'
#' @examples
#' # Standard normal confounder
#' causalsim_covar("normal", role = "confounder", mean = 0, sd = 1)
#'
#' # Binary effect modifier
#' causalsim_covar("binary", role = "effect_modifier", prob = 0.4)
#'
#' # Variable that both confounds and moderates the effect
#' causalsim_covar("normal", role = c("confounder", "effect_modifier"))
#'
#' @export
causalsim_covar <- function(dist = "normal", role = "confounder", ...) {
  dist <- match.arg(dist, c("normal", "binary", "uniform"))
  role <- match.arg(
    role,
    choices = c("confounder", "effect_modifier", "noise"),
    several.ok = TRUE
  )
  structure(
    list(dist = dist, role = role, params = list(...)),
    class = "causalsim_covar"
  )
}

#' @export
print.causalsim_covar <- function(x, ...) {
  params_str <- if (length(x$params) > 0L) {
    paste(names(x$params), x$params, sep = "=", collapse = ", ")
  } else {
    ""
  }
  dist_str <- if (nchar(params_str) > 0L) {
    sprintf("%s(%s)", x$dist, params_str)
  } else {
    x$dist
  }
  cat(sprintf("<causalsim_covar>  dist: %-20s  role: [%s]\n",
              dist_str, paste(x$role, collapse = ", ")))
  invisible(x)
}

test_that("causalsim_covar returns correct S3 class", {
  cv <- causalsim_covar("normal", role = "confounder")
  expect_s3_class(cv, "causalsim_covar")
})

test_that("causalsim_covar stores dist, role, and params", {
  cv <- causalsim_covar("binary", role = "effect_modifier", prob = 0.3)
  expect_equal(cv$dist, "binary")
  expect_equal(cv$role, "effect_modifier")
  expect_equal(cv$params$prob, 0.3)
})

test_that("causalsim_covar accepts multiple roles", {
  cv <- causalsim_covar("normal", role = c("confounder", "effect_modifier"))
  expect_equal(cv$role, c("confounder", "effect_modifier"))
})

test_that("causalsim_covar rejects invalid dist", {
  expect_error(causalsim_covar("gamma"), "should be one of")
})

test_that("causalsim_covar rejects invalid role", {
  expect_error(causalsim_covar("normal", role = "collider"), "should be one of")
})

test_that("print.causalsim_covar runs without error", {
  cv <- causalsim_covar("normal", role = "confounder", mean = 0, sd = 1)
  expect_output(print(cv), "causalsim_covar")
  expect_invisible(print(cv))
})

# ── Basic construction ────────────────────────────────────────────────────────

test_that("causalsim_dgp returns correct S3 class", {
  dgp <- causalsim_dgp(n = 100, effect = 1)
  expect_s3_class(dgp, "causalsim_dgp")
})

test_that("scalar effect is normalized to a function", {
  dgp <- causalsim_dgp(n = 100, effect = 3)
  expect_true(is.function(dgp$effect_fn))
})

test_that("heterogeneous is FALSE for scalar effect", {
  dgp <- causalsim_dgp(n = 100, effect = 2)
  expect_false(dgp$heterogeneous)
})

test_that("heterogeneous is TRUE for function effect", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1, effect = function(W) 1 + W)
  expect_true(dgp$heterogeneous)
})

test_that("n is stored as integer", {
  dgp <- causalsim_dgp(n = 200.0, effect = 1)
  expect_type(dgp$n, "integer")
  expect_equal(dgp$n, 200L)
})

test_that("mc_draws is stored as integer", {
  dgp <- causalsim_dgp(n = 100, effect = 1, mc_draws = 5000)
  expect_type(dgp$mc_draws, "integer")
})

# ── True ATE ──────────────────────────────────────────────────────────────────

test_that("true_ate is exact for scalar effect", {
  dgp <- causalsim_dgp(n = 100, effect = 2.5)
  expect_equal(dgp$true_ate, 2.5)
})

test_that("true_ate is approx correct for zero-mean heterogeneous effect", {
  # E[2 + 1.5 * W] = 2 when W ~ N(0, 1)
  set.seed(42)
  dgp <- causalsim_dgp(
    n            = 100,
    n_confounders = 1,
    effect       = function(W) 2 + 1.5 * W,
    mc_draws     = 100000L
  )
  expect_equal(dgp$true_ate, 2, tolerance = 0.02)
})

# ── sigma ─────────────────────────────────────────────────────────────────────

test_that("sigma is stored on DGP object", {
  dgp <- causalsim_dgp(n = 100, effect = 1, sigma = 2.5)
  expect_equal(dgp$sigma, 2.5)
})

test_that("negative sigma throws error", {
  expect_error(causalsim_dgp(n = 100, effect = 1, sigma = -1), "`sigma`")
})

test_that("zero sigma throws error", {
  expect_error(causalsim_dgp(n = 100, effect = 1, sigma = 0), "`sigma`")
})

# ── propensity ────────────────────────────────────────────────────────────────

test_that("propensity preset string is normalized to a function", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1,
                       propensity = "moderate", effect = 1)
  expect_true(is.function(dgp$propensity_fn))
})

test_that("propensity scalar is normalized to a function", {
  dgp <- causalsim_dgp(n = 100, propensity = 0.3, effect = 1)
  expect_true(is.function(dgp$propensity_fn))
})

test_that("propensity function returning value > 1 errors at construction", {
  expect_error(
    causalsim_dgp(
      n = 100, n_confounders = 1,
      propensity = function(W) 1.5,
      effect     = 1
    ),
    "\\[0, 1\\]"
  )
})

test_that("propensity function returning negative value errors", {
  expect_error(
    causalsim_dgp(
      n = 100, n_confounders = 1,
      propensity = function(W) -0.1,
      effect     = 1
    ),
    "\\[0, 1\\]"
  )
})

test_that("propensity referencing undefined covariate errors", {
  expect_error(
    causalsim_dgp(
      n = 100, n_confounders = 1,
      propensity = function(W, X) plogis(W + X),
      effect     = 1
    ),
    "not defined in the DGP"
  )
})

# ── baseline ──────────────────────────────────────────────────────────────────

test_that("baseline scalar is normalized to a function", {
  dgp <- causalsim_dgp(n = 100, baseline = 2, effect = 1)
  expect_true(is.function(dgp$baseline_fn))
})

test_that("baseline preset string is normalized to a function", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1,
                       baseline = "high", effect = 1)
  expect_true(is.function(dgp$baseline_fn))
})

# ── Covariate spec: shorthand (Option B) ──────────────────────────────────────

test_that("n_confounders = 1 generates covariate named W", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1, effect = 1)
  expect_named(dgp$covar_spec, "W")
  expect_equal(dgp$covar_spec$W$role, "confounder")
})

test_that("n_confounders = 2 generates W1 and W2", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 2, effect = 1)
  expect_named(dgp$covar_spec, c("W1", "W2"))
})

test_that("n_noise = 1 generates covariate named X", {
  dgp <- causalsim_dgp(n = 100, n_noise = 1, effect = 1)
  expect_named(dgp$covar_spec, "X")
  expect_equal(dgp$covar_spec$X$role, "noise")
})

test_that("n_effect_modifiers = 1 generates covariate named V", {
  dgp <- causalsim_dgp(n = 100, n_effect_modifiers = 1,
                       effect = function(V) 1 + V)
  expect_named(dgp$covar_spec, "V")
})

test_that("declaring an unused effect_modifier warns", {
  expect_warning(
    causalsim_dgp(n = 100, n_effect_modifiers = 1, effect = 1),
    "not used by `effect`"
  )
})

test_that("an effect_modifier used by `effect` does not warn", {
  expect_no_warning(
    causalsim_dgp(n = 100, n_effect_modifiers = 1, effect = function(V) 1 + V)
  )
})

# ── Covariate spec: explicit list (Option A) ──────────────────────────────────

test_that("explicit covariates list is stored on DGP", {
  dgp <- causalsim_dgp(
    n = 100,
    covariates = list(
      W = causalsim_covar("normal", role = "confounder"),
      V = causalsim_covar("binary", role = "effect_modifier", prob = 0.4)
    ),
    effect = function(V) 1 + V
  )
  expect_named(dgp$covar_spec, c("W", "V"))
  expect_s3_class(dgp$covar_spec$W, "causalsim_covar")
})

# ── Covariate spec: mixed (Option C) ─────────────────────────────────────────

test_that("shorthand and explicit covariates merge correctly", {
  dgp <- causalsim_dgp(
    n = 100,
    n_confounders = 1,
    covariates    = list(Z = causalsim_covar("normal", role = "noise")),
    effect        = 1
  )
  expect_named(dgp$covar_spec, c("W", "Z"))
})

test_that("name collision between auto and explicit covariates errors", {
  expect_error(
    causalsim_dgp(
      n = 100,
      n_confounders = 1,
      covariates    = list(W = causalsim_covar("binary", role = "confounder")),
      effect        = 1
    ),
    "appear in both"
  )
})

# ── Effect function validation ────────────────────────────────────────────────

test_that("effect referencing undefined covariate errors at construction", {
  expect_error(
    causalsim_dgp(n = 100, n_confounders = 1,
                  effect = function(W, X) W + X),
    "not defined in the DGP"
  )
})

test_that("effect referencing defined covariates does not error", {
  expect_no_error(
    causalsim_dgp(n = 100, n_confounders = 1, effect = function(W) 2 + W)
  )
})

# ── Input validation ──────────────────────────────────────────────────────────

test_that("negative n throws error", {
  expect_error(causalsim_dgp(n = -1, effect = 1), "`n`")
})

test_that("zero n throws error", {
  expect_error(causalsim_dgp(n = 0, effect = 1), "`n`")
})

test_that("non-numeric n throws error", {
  expect_error(causalsim_dgp(n = "100", effect = 1), "`n`")
})

test_that("vector effect throws error", {
  expect_error(causalsim_dgp(n = 100, effect = c(1, 2)), "`effect`")
})

test_that("non-numeric non-function effect throws error", {
  expect_error(causalsim_dgp(n = 100, effect = "large"), "preset")
})

test_that("unnamed covariates list throws error", {
  expect_error(
    causalsim_dgp(n = 100, covariates = list(causalsim_covar("normal"))),
    "must be named"
  )
})

test_that("non-causalsim_covar element in covariates list throws error", {
  expect_error(
    causalsim_dgp(n = 100, covariates = list(W = list(dist = "normal"))),
    "causalsim_covar\\(\\) objects"
  )
})

# ── Print method ──────────────────────────────────────────────────────────────

test_that("print method runs without error and returns invisibly", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1, effect = 1)
  expect_invisible(print(dgp))
  expect_output(print(dgp), "causalsim_dgp")
  expect_output(print(dgp), "true ATE")
  expect_output(print(dgp), "W")
  expect_output(print(dgp), "sigma")
})

#' Create a causal data generating process
#'
#' Defines a causal DGP with known ground truth. Covariates can be specified
#' via shorthand count arguments (Option B), an explicit named list of
#' [causalsim_covar()] objects (Option A), or both combined.
#'
#' @param n Positive integer. Sample size for each simulated dataset.
#' @param effect Numeric scalar or function. A scalar specifies a constant
#'   (homogeneous) treatment effect; ATE = CATE everywhere. A function should
#'   accept named arguments matching covariate names defined in the DGP and
#'   return a numeric vector of individual-level causal effects (CATE).
#'   See Details.
#' @param propensity Numeric scalar, preset string, or function. Treatment
#'   assignment probability. A scalar (e.g. `0.5`) gives a constant propensity
#'   (randomized trial). Preset strings `"low"`, `"moderate"`, `"high"` generate
#'   a logistic propensity over confounders with coefficients 0.25, 0.5, and
#'   1.0 respectively. A function follows the same named-argument convention as
#'   `effect` and must return values in \[0, 1\]. Defaults to `"moderate"`.
#' @param baseline Numeric scalar, preset string, or function. Mean potential
#'   outcome under control, `E[Y(0) | W]`. Preset strings follow the same
#'   levels as `propensity` and apply a linear combination of confounders.
#'   Defaults to `"moderate"`, so that confounders declared via `n_confounders`
#'   (or `role = "confounder"`) enter both the treatment and outcome models and
#'   therefore actually induce confounding bias. With no confounders present,
#'   any preset baseline resolves to `0`. Set a numeric scalar (e.g. `0`) for a
#'   constant baseline that ignores covariates.
#' @param sigma Positive numeric. Standard deviation of the outcome noise term.
#'   Default `1`.
#' @param covariates Named list of [causalsim_covar()] objects (Option A /
#'   explicit path).
#'   Each name becomes the column name in generated data and the argument name
#'   expected by `effect`, `propensity`, and `baseline` functions. Merged with
#'   any auto-generated covariates; name collisions error.
#' @param n_confounders Non-negative integer. Standard normal confounders
#'   auto-generated as `W` (single) or `W1, W2, ...` (multiple).
#' @param n_effect_modifiers Non-negative integer. Auto-generates standard
#'   normal effect modifiers as `V` or `V1, V2, ...`.
#' @param n_noise Non-negative integer. Auto-generates standard normal noise
#'   covariates as `X` or `X1, X2, ...`.
#' @param mc_draws Positive integer. Monte Carlo draws for true ATE
#'   approximation. Default `10000L`. Ignored for scalar `effect`.
#'
#' @details
#' ## Structural model
#'
#' ```
#' W ~ covariate_spec
#' A ~ Bernoulli(propensity(W))
#' Y  = baseline(W) + effect(W) * A + N(0, sigma^2)
#' ```
#'
#' `A` is used for treatment throughout to avoid collision with R's built-in
#' `T` alias.
#'
#' ## Function calling convention
#'
#' The `effect`, `propensity`, and `baseline` functions are called with named
#' arguments matching covariate names, not a data frame. Write:
#'
#' ```r
#' effect = function(W) 2 + 1.5 * W
#' propensity = function(W1, W2) plogis(0.3 * W1 + 0.5 * W2)
#' ```
#'
#' Every argument name is validated against the DGP's covariate spec at
#' construction time, so mismatches surface immediately rather than at draw
#' time. The propensity function is additionally evaluated on a small test draw
#' to confirm it returns values in \[0, 1\].
#'
#' ## True ATE
#'
#' For scalar `effect` the true ATE is exact. For function `effect` it is
#' approximated via Monte Carlo over `mc_draws` draws from the covariate
#' distribution.
#'
#' @return An S3 object of class `causalsim_dgp` with components:
#' \describe{
#'   \item{`n`}{Sample size (integer)}
#'   \item{`covar_spec`}{Named list of [causalsim_covar()] objects}
#'   \item{`effect_fn`}{Normalized effect function}
#'   \item{`propensity_fn`}{Normalized propensity function}
#'   \item{`baseline_fn`}{Normalized baseline function}
#'   \item{`sigma`}{Outcome noise standard deviation}
#'   \item{`true_ate`}{True ATE: exact for scalar, MC approximation otherwise}
#'   \item{`heterogeneous`}{Logical; `TRUE` if `effect` was a function}
#'   \item{`mc_draws`}{Monte Carlo draws used (integer)}
#' }
#'
#' @examples
#' # Minimal: one confounder, constant effect, moderate confounding
#' dgp <- causalsim_dgp(n = 500, n_confounders = 1, effect = 2)
#' dgp
#'
#' # Heterogeneous effect, explicit covariate spec
#' dgp2 <- causalsim_dgp(
#'   n = 500,
#'   covariates = list(
#'     W = causalsim_covar("normal", role = "confounder"),
#'     V = causalsim_covar("binary", role = "effect_modifier", prob = 0.4)
#'   ),
#'   effect = function(V) 2 + 1.5 * V,
#'   propensity = function(W) plogis(0.5 * W),
#'   baseline = function(W) 1.5 * W
#' )
#'
#' # Mixed: shorthand confounders + explicit noise covariate + RCT propensity
#' dgp3 <- causalsim_dgp(
#'   n = 1000,
#'   n_confounders = 2,
#'   covariates = list(X = causalsim_covar("normal", role = "noise")),
#'   effect = 1,
#'   propensity = 0.5
#' )
#'
#' @export
causalsim_dgp <- function(
  n,
  effect = 1,
  propensity = "moderate",
  baseline = "moderate",
  sigma = 1,
  covariates = list(),
  n_confounders = 0L,
  n_effect_modifiers = 0L,
  n_noise = 0L,
  mc_draws = 10000L
) {
  n <- .validate_positive(n, "n", as_int = TRUE)
  mc_draws <- .validate_positive(mc_draws, "mc_draws", as_int = TRUE)
  sigma <- .validate_positive(sigma, "sigma")
  n_confounders <- .validate_nonneg(
    n_confounders, "n_confounders", as_int = TRUE
  )
  n_effect_modifiers <- .validate_nonneg(
    n_effect_modifiers, "n_effect_modifiers", as_int = TRUE
  )
  n_noise <- .validate_nonneg(n_noise, "n_noise", as_int = TRUE)

  covar_spec <- .build_covariate_spec(
    n_confounders, n_effect_modifiers, n_noise, covariates
  )

  heterogeneous <- is.function(effect)
  effect_fn <- .normalize_fn(effect, "effect", covar_spec)
  propensity_fn <- .normalize_fn(propensity, "propensity", covar_spec)
  baseline_fn <- .normalize_fn(baseline, "baseline", covar_spec)

  .validate_fn_args(effect_fn, names(covar_spec), "effect")
  .validate_fn_args(propensity_fn, names(covar_spec), "propensity")
  .validate_fn_args(baseline_fn, names(covar_spec), "baseline")
  .validate_propensity_fn(propensity_fn, covar_spec)
  .warn_inert_effect_modifiers(effect, covar_spec)

  true_ate <- .mc_ate(effect_fn, covar_spec, mc_draws, heterogeneous)

  structure(
    list(
      n = n,
      covar_spec = covar_spec,
      effect_fn = effect_fn,
      propensity_fn = propensity_fn,
      baseline_fn = baseline_fn,
      sigma = sigma,
      true_ate = true_ate,
      heterogeneous = heterogeneous,
      mc_draws = mc_draws,
      params = list(
        n = n,
        effect = effect,
        propensity = propensity,
        baseline = baseline,
        sigma = sigma,
        covariates = covariates,
        n_confounders = n_confounders,
        n_effect_modifiers = n_effect_modifiers,
        n_noise = n_noise,
        mc_draws = mc_draws
      )
    ),
    class = "causalsim_dgp"
  )
}

#' @export
print.causalsim_dgp <- function(x, ...) {
  cat("<causalsim_dgp>\n")
  cat(sprintf("  n            : %d\n", x$n))
  cat(sprintf("  true ATE     : %.4f\n", x$true_ate))
  cat(sprintf("  heterogeneous: %s\n", x$heterogeneous))
  cat(sprintf("  sigma        : %.2f\n", x$sigma))
  if (length(x$covar_spec) > 0L) {
    cat("  covariates   :\n")
    for (nm in names(x$covar_spec)) {
      cv <- x$covar_spec[[nm]]
      cat(sprintf("    %-6s  %s  [%s]\n",
                  nm, cv$dist, paste(cv$role, collapse = "+")))
    }
  }
  invisible(x)
}

# ── Return type and shape ─────────────────────────────────────────────────────

test_that("causalsim_draw returns a data frame", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1, effect = 1)
  expect_s3_class(causalsim_draw(dgp), "data.frame")
})

test_that("causalsim_draw returns dgp$n rows", {
  dgp <- causalsim_dgp(n = 250, n_confounders = 1, effect = 1)
  expect_equal(nrow(causalsim_draw(dgp)), 250L)
})

test_that("output has covariate, treatment, outcome, and meta columns", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1, effect = 1)
  d   <- causalsim_draw(dgp)
  expect_true(all(c("W", "A", "Y", ".tau", ".p") %in% names(d)))
})

test_that("multi-covariate DGP produces correctly named columns", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 2,
                       n_noise = 1, effect = 1)
  d   <- causalsim_draw(dgp)
  expect_true(all(c("W1", "W2", "X", "A", "Y") %in% names(d)))
})

# ── Column types and ranges ───────────────────────────────────────────────────

test_that("treatment A is binary (0 or 1)", {
  dgp <- causalsim_dgp(n = 200, n_confounders = 1, effect = 1)
  d   <- causalsim_draw(dgp)
  expect_true(all(d$A %in% c(0L, 1L)))
})

test_that(".p values are in [0, 1]", {
  dgp <- causalsim_dgp(n = 200, n_confounders = 1, effect = 1)
  d   <- causalsim_draw(dgp)
  expect_true(all(d$.p >= 0 & d$.p <= 1))
})

test_that("Y is numeric", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1, effect = 1)
  expect_type(causalsim_draw(dgp)$Y, "double")
})

# ── Reproducibility ───────────────────────────────────────────────────────────

test_that("same seed produces identical draws", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1, effect = 1)
  d1  <- causalsim_draw(dgp, seed = 42)
  d2  <- causalsim_draw(dgp, seed = 42)
  expect_identical(d1, d2)
})

test_that("different seeds produce different draws", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1, effect = 1)
  d1  <- causalsim_draw(dgp, seed = 1)
  d2  <- causalsim_draw(dgp, seed = 2)
  expect_false(isTRUE(all.equal(d1$Y, d2$Y)))
})

# ── Ground truth columns ──────────────────────────────────────────────────────

test_that(".tau is constant for scalar effect", {
  dgp <- causalsim_dgp(n = 100, n_confounders = 1, effect = 3)
  d   <- causalsim_draw(dgp, seed = 1)
  expect_true(all(d$.tau == 3))
})

test_that(".tau varies for heterogeneous effect", {
  dgp <- causalsim_dgp(n = 200, n_confounders = 1,
                       effect = function(W) 2 + W)
  d   <- causalsim_draw(dgp, seed = 1)
  expect_gt(var(d$.tau), 0)
})

test_that("mean .tau is close to true_ate for large draw", {
  # E[2 + W] = 2 when W ~ N(0,1)
  set.seed(7)
  dgp <- causalsim_dgp(n = 10000, n_confounders = 1,
                       effect = function(W) 2 + W)
  d   <- causalsim_draw(dgp)
  expect_equal(mean(d$.tau), dgp$true_ate, tolerance = 0.05)
})

# ── Confounding behaviour ─────────────────────────────────────────────────────

test_that("a default confounder actually induces naive bias", {
  # With default propensity + baseline presets, a confounder must enter both
  # the treatment and outcome models, so a naive difference in means is biased
  # away from the true ATE.
  dgp <- causalsim_dgp(n = 50000, n_confounders = 1, effect = 2)
  d   <- causalsim_draw(dgp, seed = 1)
  naive <- mean(d$Y[d$A == 1]) - mean(d$Y[d$A == 0])
  expect_gt(abs(naive - dgp$true_ate), 0.05)
})

test_that("an RCT (scalar propensity) leaves the naive estimate unbiased", {
  # Randomized treatment: even with a preset baseline the confounder cannot
  # induce bias, because assignment does not depend on it.
  dgp <- causalsim_dgp(n = 50000, n_confounders = 1, effect = 2,
                       propensity = 0.5)
  d   <- causalsim_draw(dgp, seed = 1)
  naive <- mean(d$Y[d$A == 1]) - mean(d$Y[d$A == 0])
  expect_lt(abs(naive - dgp$true_ate), 0.05)
})

# ── No-covariate DGP ─────────────────────────────────────────────────────────

test_that("draw works with no covariates and scalar functions", {
  dgp <- causalsim_dgp(n = 100, effect = 2, propensity = 0.5, baseline = 0)
  d   <- causalsim_draw(dgp, seed = 1)
  expect_equal(nrow(d), 100L)
  expect_true(all(d$.tau == 2))
  expect_true(all(d$.p  == 0.5))
})

# ── Input validation ──────────────────────────────────────────────────────────

test_that("non-causalsim_dgp input errors", {
  expect_error(causalsim_draw(list()), "causalsim_dgp")
})

#' causalsim: Simulation-Ready Causal Data Generating Processes
#'
#' Construct, simulate, and evaluate causal data generating processes (DGPs)
#' with known ground truth. Covariate roles (confounder, effect modifier,
#' noise) and heterogeneous treatment effects are first-class concepts, and
#' estimator performance is summarised with bias, RMSE, confidence-interval
#' coverage, and power. Results are returned as plain data frames for further
#' analysis or plotting.
#'
#' @keywords internal
"_PACKAGE"

# Null-coalescing: return x unless it is NULL, then return y
`%||%` <- function(x, y) if (is.null(x)) y else x

# ── Input validation ──────────────────────────────────────────────────────────

.validate_positive <- function(x, name, as_int = FALSE) {
  if (!is.numeric(x) || length(x) != 1L || !is.finite(x) || x <= 0) {
    stop(sprintf("`%s` must be a finite positive number.", name), call. = FALSE)
  }
  if (as_int) as.integer(x) else x
}

.validate_nonneg <- function(x, name, as_int = FALSE) {
  if (!is.numeric(x) || length(x) != 1L || !is.finite(x) || x < 0) {
    stop(
      sprintf("`%s` must be a finite non-negative number.", name),
      call. = FALSE
    )
  }
  if (as_int) as.integer(x) else x
}

# ── Covariate spec construction ───────────────────────────────────────────────

# Merge shorthand count arguments with an explicit named covariate list.
.build_covariate_spec <- function(n_confounders, n_effect_modifiers,
                                  n_noise, covariates) {
  auto <- c(
    .auto_covariates(n_confounders, "confounder", "W"),
    .auto_covariates(n_effect_modifiers, "effect_modifier", "V"),
    .auto_covariates(n_noise, "noise", "X")
  )

  if (!is.list(covariates)) {
    stop(
      "`covariates` must be a named list of causalsim_covar() objects.",
      call. = FALSE
    )
  }
  if (length(covariates) > 0L) {
    if (is.null(names(covariates)) || any(names(covariates) == "")) {
      stop("Every element of `covariates` must be named.", call. = FALSE)
    }
    not_covar <- !vapply(
      covariates, inherits, logical(1L), what = "causalsim_covar"
    )
    if (any(not_covar)) {
      stop(
        "All elements of `covariates` must be causalsim_covar() objects.",
        call. = FALSE
      )
    }
    overlap <- intersect(names(auto), names(covariates))
    if (length(overlap) > 0L) {
      stop(
        sprintf(
          paste0(
            "Covariate name(s) appear in both auto-generated",
            " and explicit spec: %s"
          ),
          paste(overlap, collapse = ", ")
        ),
        call. = FALSE
      )
    }
  }

  c(auto, covariates)
}

# Naming convention: single item → bare prefix (W, Z, V, X);
# multiple items → numbered suffix (W1, W2, ...).
.auto_covariates <- function(n, role, prefix) {
  if (n == 0L) return(list())
  nms  <- if (n == 1L) prefix else paste0(prefix, seq_len(n))
  covs <- replicate(
    n, causalsim_covar(dist = "normal", role = role), simplify = FALSE
  )
  stats::setNames(covs, nms)
}

# ── Covariate data generation ─────────────────────────────────────────────────

# Generate an n-row data frame from a named covariate spec list.
.generate_covariate_data <- function(covar_spec, n) {
  if (length(covar_spec) == 0L) return(data.frame())
  cols <- lapply(covar_spec, .draw_one_covariate, n = n)
  as.data.frame(cols, check.names = FALSE)
}

# Draw n observations for a single causalsim_covar object.
.draw_one_covariate <- function(cv, n) {
  p <- cv$params
  switch(cv$dist,
    normal = stats::rnorm(n, mean = p$mean %||% 0, sd = p$sd %||% 1),
    binary = stats::rbinom(n, size = 1L, prob = p$prob %||% 0.5),
    uniform = stats::runif(n, min = p$min %||% 0, max = p$max %||% 1),
    stop(sprintf("Unknown distribution: '%s'", cv$dist), call. = FALSE)
  )
}

# ── Function normalization ────────────────────────────────────────────────────

# Preset table: maps (type, level) → linear coefficient on confounder columns.
# Propensity uses plogis(); baseline uses a linear sum.
.fn_presets <- list(
  propensity = c(low = 0.25, moderate = 0.5, high = 1.0),
  baseline = c(low = 0.25, moderate = 0.5, high = 1.0)
)

# Resolve a preset string to a closure over the covariate spec.
# The returned function uses `...` and expects to be called via .apply_effect.
.resolve_preset <- function(x, type, covar_spec) {
  table <- .fn_presets[[type]]
  if (is.null(table)) {
    stop(
      sprintf(
        "No presets defined for `%s`. Use a scalar or function.", type
      ),
      call. = FALSE
    )
  }
  if (!x %in% names(table)) {
    stop(
      sprintf(
        "`%s` preset '%s' not recognized. Choose from: %s.",
        type, x, paste(names(table), collapse = ", ")
      ),
      call. = FALSE
    )
  }

  coeff <- unname(table[[x]])
  confounder_nms <- names(
    Filter(function(cv) "confounder" %in% cv$role, covar_spec)
  )

  # No confounders: preset has no structural effect
  if (length(confounder_nms) == 0L) {
    default_val <- if (type == "propensity") 0.5 else 0
    return(function(...) default_val)
  }

  # Return closure capturing coeff and confounder names
  coeff_ <- coeff
  confounder_nms_ <- confounder_nms
  if (type == "propensity") {
    function(...) {
      args <- list(...)
      terms <- lapply(confounder_nms_, function(nm) coeff_ * args[[nm]])
      lin <- Reduce(`+`, terms)
      stats::plogis(lin)
    }
  } else {
    function(...) {
      args <- list(...)
      Reduce(`+`, lapply(confounder_nms_, function(nm) coeff_ * args[[nm]]))
    }
  }
}

# Normalize effect/propensity/baseline to a callable function.
# Accepts a scalar (constant), a preset string, or a user-supplied function.
.normalize_fn <- function(x, type, covar_spec) {
  if (is.numeric(x) && length(x) == 1L && is.finite(x)) {
    val <- x
    return(function(...) val)
  }
  if (is.character(x) && length(x) == 1L) {
    return(.resolve_preset(x, type, covar_spec))
  }
  if (is.function(x)) {
    return(x)
  }
  stop(
    sprintf("`%s` must be a scalar, preset string, or function.", type),
    call. = FALSE
  )
}

# ── Function application ──────────────────────────────────────────────────────

# Apply a DGP function to a covariate data frame, always returning a
# length-n vector.
#
# Dispatch rules:
#   no-arg function  (formals = "")       -> scalar, repeat n times
#   ...-only function (formals = "...")   -> do.call with full covariate df
#   named-arg function                    -> do.call with matching columns only
#
# The `n` argument is used when covariate_df has 0 rows (no-covariate DGPs).
.apply_effect <- function(effect_fn, covariate_df, n = nrow(covariate_df)) {
  fn_formals <- names(formals(effect_fn))
  has_dots <- "..." %in% fn_formals
  named_args <- setdiff(fn_formals, "...")

  result <- if (!has_dots && length(named_args) == 0L) {
    effect_fn()
  } else if (has_dots && length(named_args) == 0L) {
    do.call(effect_fn, as.list(covariate_df))
  } else {
    do.call(effect_fn, as.list(covariate_df[named_args]))
  }

  # Recycle scalars to length n so callers always get a vector
  if (length(result) == 1L && n > 1L) rep(result, n) else result
}

# ── Argument validation ───────────────────────────────────────────────────────

# Validate that every named argument of fn exists in covariate_names.
# Skips ...-only and no-arg functions (no names to validate).
.validate_fn_args <- function(fn, covariate_names, arg_label) {
  named_args <- setdiff(names(formals(fn)), "...")
  if (length(named_args) == 0L) return(invisible(NULL))
  missing_vars <- setdiff(named_args, covariate_names)
  if (length(missing_vars) > 0L) {
    stop(
      sprintf(
        "`%s` function references covariate(s) not defined in the DGP: %s",
        arg_label, paste(missing_vars, collapse = ", ")
      ),
      call. = FALSE
    )
  }
}

# Warn when a covariate is declared an effect modifier but the `effect` argument
# never references it, so it silently has no bearing on the treatment effect.
# `effect` is the ORIGINAL user argument (scalar/preset/function), not the
# normalized closure: a normalized scalar gains a `...` formal and would be
# indistinguishable from a user function that consumes all covariates.
.warn_inert_effect_modifiers <- function(effect, covar_spec) {
  modifier_nms <- names(
    Filter(function(cv) "effect_modifier" %in% cv$role, covar_spec)
  )
  if (length(modifier_nms) == 0L) return(invisible(NULL))

  if (is.function(effect)) {
    fn_formals <- names(formals(effect))
    # A `...`-consuming effect may use any covariate; cannot flag as inert.
    if ("..." %in% fn_formals) return(invisible(NULL))
    unused <- setdiff(modifier_nms, fn_formals)
  } else {
    # Scalar or preset effect is constant across covariates.
    unused <- modifier_nms
  }

  if (length(unused) > 0L) {
    warning(
      sprintf(
        paste0(
          "effect_modifier covariate(s) not used by `effect`: %s. They will ",
          "not modify the treatment effect. Reference them in a function ",
          "passed to `effect`, e.g. effect = function(%s) 2 + %s."
        ),
        paste(unused, collapse = ", "), unused[1L], unused[1L]
      ),
      call. = FALSE
    )
  }
  invisible(NULL)
}

# Validate that the propensity function returns values strictly in [0, 1]
# by evaluating it on a small test draw at construction time.
.validate_propensity_fn <- function(propensity_fn, covar_spec,
                                    n_check = 50L) {
  test_df <- if (length(covar_spec) > 0L) {
    .generate_covariate_data(covar_spec, n = n_check)
  } else {
    data.frame()
  }

  probs <- tryCatch(
    .apply_effect(propensity_fn, test_df, n = n_check),
    error = function(e) {
      stop(
        "Failed to evaluate `propensity` function: ", conditionMessage(e),
        call. = FALSE
      )
    }
  )

  if (length(probs) > 0L &&
        (any(!is.finite(probs)) || any(probs < 0) || any(probs > 1))) {
    stop(
      sprintf(
        paste0("`propensity` function must return values in [0, 1]. ",
               "Got range [%.3g, %.3g]."),
        min(probs, na.rm = TRUE), max(probs, na.rm = TRUE)
      ),
      call. = FALSE
    )
  }
}

# ── Monte Carlo ATE ───────────────────────────────────────────────────────────

# Approximate the true ATE via Monte Carlo over the covariate distribution.
# For non-heterogeneous (scalar) effects, returns the exact value immediately.
.mc_ate <- function(effect_fn, covar_spec, mc_draws, heterogeneous) {
  if (!heterogeneous) return(effect_fn())
  covariate_df <- .generate_covariate_data(covar_spec, n = mc_draws)
  mean(.apply_effect(effect_fn, covariate_df, n = mc_draws))
}

# ── Estimator evaluation ──────────────────────────────────────────────────────

# Call a user estimator on one dataset and return a named numeric vector.
# Accepts a named numeric vector, a named all-numeric list, or a one-row data
# frame. Errors with a clear message if 'estimate' is not present.
.call_estimator <- function(estimator, data) {
  result <- estimator(data)
  if (is.data.frame(result)) {
    if (nrow(result) != 1L) {
      stop(
        "estimator must return a named numeric vector, named list, or ",
        "one-row data frame.", call. = FALSE
      )
    }
    result <- unlist(result[1L, ], use.names = TRUE)
  }
  if (is.list(result)) {
    all_scalar_numeric <- !is.null(names(result)) &&
      all(vapply(result, function(x) is.numeric(x) && length(x) == 1L,
                 logical(1L)))
    if (!all_scalar_numeric) {
      stop(
        "estimator returned a list, but not all elements are named numeric ",
        "scalars. Use c() or return a one-row data frame instead.",
        call. = FALSE
      )
    }
    result <- stats::setNames(as.numeric(result), names(result))
  }
  if (!is.numeric(result) || is.null(names(result))) {
    stop(
      sprintf(
        "estimator must return a named numeric vector, named list, or ",
        "one-row data frame; got %s.", class(result)[1L]
      ),
      call. = FALSE
    )
  }
  if (!"estimate" %in% names(result)) {
    stop(
      "estimator output must include a field named 'estimate'.",
      call. = FALSE
    )
  }
  result
}

# Run `reps` draws from `dgp`, apply `estimator` to each, and return a
# data frame with one row per replication.
.collect_draws <- function(dgp, estimator, reps) {
  rows <- vector("list", reps)
  for (i in seq_len(reps)) {
    data <- causalsim_draw(dgp)
    est <- .call_estimator(estimator, data)
    rows[[i]] <- as.list(est)
  }
  # Bind rows; missing fields (e.g. ci_lower absent in some reps) become NA
  all_nms <- unique(unlist(lapply(rows, names)))
  out <- as.data.frame(
    do.call(rbind, lapply(rows, function(r) {
      r[setdiff(all_nms, names(r))] <- NA_real_
      as.data.frame(r[all_nms])
    })),
    stringsAsFactors = FALSE
  )
  out
}

# Compute requested metrics from a draws data frame and the true ATE.
# Returns a tidy data frame: metric | value | se
.compute_metrics <- function(draws, true_ate, metrics) {
  est <- draws[["estimate"]]
  errs <- est - true_ate

  rows <- lapply(metrics, function(m) {
    switch(m,
      bias = {
        val <- mean(errs)
        se <- stats::sd(est) / sqrt(length(est))
        data.frame(metric = "bias", value = val, se = se,
                   stringsAsFactors = FALSE)
      },
      rmse = {
        sq <- errs^2
        val <- sqrt(mean(sq))
        # Delta method: SE(sqrt(mean(X))) ≈ SD(X) / (2*sqrt(mean(X))*sqrt(n))
        se <- if (val > 0) {
          stats::sd(sq) / (2 * val * sqrt(length(sq)))
        } else {
          0
        }
        data.frame(metric = "rmse", value = val, se = se,
                   stringsAsFactors = FALSE)
      },
      coverage = {
        covered <- draws[["ci_lower"]] <= true_ate &
                   true_ate <= draws[["ci_upper"]]
        p <- mean(covered)
        se <- sqrt(p * (1 - p) / length(covered))
        data.frame(metric = "coverage", value = p, se = se,
                   stringsAsFactors = FALSE)
      },
      power = {
        rejected <- draws[["ci_lower"]] > 0 | draws[["ci_upper"]] < 0
        p <- mean(rejected)
        se <- sqrt(p * (1 - p) / length(rejected))
        data.frame(metric = "power", value = p, se = se,
                   stringsAsFactors = FALSE)
      }
    )
  })

  do.call(rbind, rows)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  fig.width  = 7,
  fig.height = 4.5,
  out.width  = "100%"
)

## ----library------------------------------------------------------------------
library(causalsim)

## ----dgp----------------------------------------------------------------------
dgp <- causalsim_dgp(
  n = 500,
  n_confounders = 1,
  effect = 2,
  propensity = "moderate",
  baseline = "moderate"
)
dgp

## ----draw---------------------------------------------------------------------
dat <- causalsim_draw(dgp, seed = 1L)
head(dat)

## ----confounding-check--------------------------------------------------------
aggregate(W ~ A, data = dat, FUN = mean)

## ----estimators---------------------------------------------------------------
# Naive: regresses Y on A only, omits the confounder
naive_est <- function(data) {
  fit <- lm(Y ~ A, data = data)
  est <- coef(fit)[["A"]]
  se <- sqrt(vcov(fit)["A", "A"])
  c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
}

# OLS: adjusts for the observed confounder W
ols_est <- function(data) {
  fit <- lm(Y ~ A + W, data = data)
  est <- coef(fit)[["A"]]
  se <- sqrt(vcov(fit)["A", "A"])
  c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
}

## ----estimator-list, eval = FALSE---------------------------------------------
# ols_est <- function(data) {
#   fit <- lm(Y ~ A + W, data = data)
#   ci <- confint(fit)["A", ]
#   list(
#     estimate = coef(fit)["A"],
#     ci_lower = ci[1],
#     ci_upper = ci[2]
#   )
# }

## ----eval-naive---------------------------------------------------------------
eval_naive <- causalsim_eval(dgp, naive_est, reps = 300L, seed = 1L)
eval_naive

## ----eval-ols-----------------------------------------------------------------
eval_ols <- causalsim_eval(dgp, ols_est, reps = 300L, seed = 1L)
eval_ols

## ----summary-ols--------------------------------------------------------------
summary(eval_ols)

## ----plot-ols, fig.cap = "Distribution of OLS estimates over 300 replications. Solid line: true ATE. Dashed line: mean estimate."----
plot(eval_ols)

## ----grid-n-------------------------------------------------------------------
grid_n <- causalsim_eval_grid(
  dgp = dgp,
  estimator = ols_est,
  vary = list(n = c(100L, 250L, 500L, 1000L)),
  reps = 300L,
  metrics = c("bias", "rmse"),
  seed = 1L
)
grid_n

## ----grid-confounding---------------------------------------------------------
conf_levels <- list(propensity = c("low", "moderate", "high"))

grid_naive <- causalsim_eval_grid(dgp, naive_est,
                             vary = conf_levels,
                             reps = 300L,
                             metrics = "bias",
                             seed = 1L)

grid_ols <- causalsim_eval_grid(dgp, ols_est,
                           vary = conf_levels,
                           reps = 300L,
                           metrics = "bias",
                           seed = 1L)

comparison <- rbind(
  cbind(estimator = "naive", grid_naive$results),
  cbind(estimator = "ols",   grid_ols$results)
)
comparison <- comparison[order(comparison$propensity, comparison$estimator), ]
rownames(comparison) <- NULL
comparison

## ----het-dgp------------------------------------------------------------------
het_dgp <- causalsim_dgp(
  n = 4000,
  covariates = list(
    W = causalsim_covar("normal", role = "confounder"),
    V = causalsim_covar("binary", role = "effect_modifier", prob = 0.5)
  ),
  effect = function(V) 2 + 3 * V,   # effect is 2 when V = 0, 5 when V = 1
  propensity = function(W) plogis(0.5 * W),
  baseline = function(W) W
)
het_dgp

## ----het-truth----------------------------------------------------------------
d <- causalsim_draw(het_dgp, seed = 1L)
tapply(d$.tau, d$V, mean)   # 2 for V = 0, 5 for V = 1

## ----het-estimators-----------------------------------------------------------
overall <- lm(Y ~ A + W, data = d)         # assumes a constant effect
interact <- lm(Y ~ A * V + W, data = d)     # allows the effect to vary with V

c(
  average = coef(overall)[["A"]],
  subgroup_v0 = coef(interact)[["A"]],
  subgroup_v1 = coef(interact)[["A"]] + coef(interact)[["A:V"]]
)

## ----session-info-------------------------------------------------------------
sessionInfo()

#' Simulate a causal dataset with known ground truth
#'
#' A convenience wrapper around [causalsim_dgp()] and [causalsim_draw()].
#' Defines a causal data generating process and returns one simulated dataset
#' in a single call.
#'
#' @inheritParams causalsim_dgp
#' @param seed Integer or `NULL`. Passed to [set.seed()] before drawing.
#'   Default `NULL`.
#'
#' @details
#' The returned data frame includes all covariate columns, the treatment
#' indicator `A`, the outcome `Y`, and two ground-truth columns:
#'
#' \describe{
#'   \item{`.tau`}{Individual-level causal effect (CATE).}
#'   \item{`.p`}{True propensity score.}
#' }
#'
#' For full control over the DGP (multiple draws, grid evaluation, estimator
#' benchmarking), use [causalsim_dgp()] and [causalsim_draw()] directly.
#'
#' @return A data frame with covariate columns, `A`, `Y`, `.tau`, and `.p`.
#'
#' @examples
#' # One confounder, constant effect of 2, moderate confounding
#' data <- causalsim(n = 500, n_confounders = 1, effect = 2, seed = 1L)
#' head(data)
#'
#' # Heterogeneous effect with an explicit covariate spec
#' data2 <- causalsim(
#'   n = 500,
#'   covariates = list(
#'     W = causalsim_covar("normal", role = "confounder"),
#'     V = causalsim_covar("binary", role = "effect_modifier", prob = 0.4)
#'   ),
#'   effect = function(V) 2 + 1.5 * V,
#'   propensity = function(W) plogis(0.5 * W),
#'   seed = 42L
#' )
#' head(data2)
#'
#' @export
causalsim <- function(
  n,
  effect = 1,
  propensity = "moderate",
  baseline = "moderate",
  sigma = 1,
  covariates = list(),
  n_confounders = 0L,
  n_effect_modifiers = 0L,
  n_noise = 0L,
  mc_draws = 10000L,
  seed = NULL
) {
  dgp <- causalsim_dgp(
    n = n,
    effect = effect,
    propensity = propensity,
    baseline = baseline,
    sigma = sigma,
    covariates = covariates,
    n_confounders = n_confounders,
    n_effect_modifiers = n_effect_modifiers,
    n_noise = n_noise,
    mc_draws = mc_draws
  )
  if (!is.null(seed)) set.seed(seed)
  causalsim_draw(dgp)
}

#' Draw a dataset from a causal DGP
#'
#' Simulates one dataset from a [causalsim_dgp()] object using the structural
#' model:
#'
#' ```
#' W ~ covariate_spec
#' A ~ Bernoulli(propensity(W))
#' Y  = baseline(W) + effect(W) * A + N(0, sigma^2)
#' ```
#'
#' @param dgp A `causalsim_dgp` object created by [causalsim_dgp()].
#' @param seed Integer or `NULL`. If non-null, passed to [set.seed()] before
#'   generating any random values, making the draw reproducible. Default `NULL`.
#'
#' @return A data frame with `dgp$n` rows and the following columns:
#' \describe{
#'   \item{Covariate columns}{One column per covariate in `dgp$covar_spec`,
#'     named to match the spec (e.g. `W`, `W1`, `Z`).}
#'   \item{`A`}{Binary treatment indicator (0/1).}
#'   \item{`Y`}{Observed outcome.}
#'   \item{`.tau`}{Individual treatment effect (CATE). Ground truth.}
#'   \item{`.p`}{Individual propensity score. Ground truth.}
#' }
#'
#' The `.tau` and `.p` columns carry individual-level ground truth and are
#' prefixed with `.` to distinguish them from observed variables.
#' `causalsim_eval()` uses them directly without re-estimation.
#'
#' @examples
#' dgp <- causalsim_dgp(n = 500, n_confounders = 1, effect = 2)
#' d   <- causalsim_draw(dgp, seed = 1)
#' head(d)
#'
#' # Reproducible draws
#' d1 <- causalsim_draw(dgp, seed = 42)
#' d2 <- causalsim_draw(dgp, seed = 42)
#' identical(d1, d2)  # TRUE
#'
#' @export
causalsim_draw <- function(dgp, seed = NULL) {
  if (!inherits(dgp, "causalsim_dgp")) {
    stop("`dgp` must be a causalsim_dgp object.", call. = FALSE)
  }
  if (!is.null(seed)) set.seed(seed)

  n <- dgp$n
  cov_df <- .generate_covariate_data(dgp$covar_spec, n)

  p <- .apply_effect(dgp$propensity_fn, cov_df, n = n)
  a <- stats::rbinom(n, size = 1L, prob = p)
  mu <- .apply_effect(dgp$baseline_fn, cov_df, n = n)
  tau <- .apply_effect(dgp$effect_fn, cov_df, n = n)
  eps <- stats::rnorm(n, mean = 0, sd = dgp$sigma)
  y <- mu + tau * a + eps

  if (length(dgp$covar_spec) > 0L) {
    data.frame(cov_df, A = a, Y = y, .tau = tau, .p = p,
               check.names = FALSE)
  } else {
    data.frame(A = a, Y = y, .tau = tau, .p = p)
  }
}

---
title: "A Complete Simulation Study with causalsim"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{A Complete Simulation Study with causalsim}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  fig.width  = 7,
  fig.height = 4.5,
  out.width  = "100%"
)
```

```{css, echo = FALSE}
.callout {
  background-color: #f0f4f8;
  border-left: 4px solid #3a7ebf;
  padding: 0.75em 1em;
  margin: 1.25em 0;
  border-radius: 0 4px 4px 0;
}

.callout-warn {
  background-color: #fff8e1;
  border-left: 4px solid #f0a500;
  padding: 0.75em 1em;
  margin: 1.25em 0;
  border-radius: 0 4px 4px 0;
}

table {
  width: 100%;
  border-collapse: collapse;
  font-size: 0.9em;
  margin: 1em 0;
}

th {
  background-color: #3a7ebf;
  color: white;
  padding: 0.5em 0.75em;
  text-align: left;
}

td {
  padding: 0.4em 0.75em;
  border-bottom: 1px solid #dde;
}

tr:nth-child(even) td {
  background-color: #f5f7fa;
}
```

## Overview

Evaluating causal estimators requires knowing the answer. `causalsim` gives you
that by making the data-generating process explicit: you specify the structural
model, the package simulates data from it, and you measure how well any estimator
recovers the truth you specified.

If you only need a single dataset with known ground truth, `causalsim()` does it
in one call:

```r
data <- causalsim(n = 500, n_confounders = 1, effect = 2, seed = 1L)
head(data)
```

The columns `.tau` and `.p` carry the ground truth for whatever analysis you run
on the data. This vignette goes further: defining a DGP explicitly, evaluating
estimator performance over many replications, and sweeping across a parameter
grid.

### What this vignette covers

| Step | Function | What it does |
|---|---|---|
| 0 | `causalsim()` | Simulate one dataset in a single call |
| 1 | `causalsim_dgp()` | Define the structural model and true ATE |
| 2 | `causalsim_draw()` | Simulate one dataset and inspect it |
| 3 | `causalsim_eval()` | Measure estimator performance over many replications |
| 4 | `causalsim_eval_grid()` | Sweep over sample sizes and confounding levels |

```{r library}
library(causalsim)
```

---

## Step 1: Define the DGP

The structural model for this study is:

$$
W \sim N(0,1), \quad
A \mid W \sim \text{Bernoulli}\!\left(\text{logistic}(0.5\,W)\right), \quad
Y = 2A + 0.5W + \varepsilon, \quad
\varepsilon \sim N(0,1)
$$

In `causalsim_dgp()` terms: one standard-normal confounder, a constant effect of
2, moderate propensity confounding (logistic coefficient 0.5), and a moderate
baseline shift.

```{r dgp}
dgp <- causalsim_dgp(
  n = 500,
  n_confounders = 1,
  effect = 2,
  propensity = "moderate",
  baseline = "moderate"
)
dgp
```

<div class="callout">
The true ATE is exact when `effect` is a scalar; no Monte Carlo approximation
is needed. For function-valued effects, `causalsim_dgp()` approximates the ATE
via 10,000 Monte Carlo draws at construction time.
</div>

---

## Step 2: Inspect a Draw

`causalsim_draw()` simulates one dataset from the DGP. The columns `.tau` and
`.p` are the individual causal effect and propensity score, diagnostic metadata
that is not available in real observational data.

```{r draw}
dat <- causalsim_draw(dgp, seed = 1L)
head(dat)
```

With moderate confounding, the treated and control groups differ on the
pre-treatment covariate:

```{r confounding-check}
aggregate(W ~ A, data = dat, FUN = mean)
```

<div class="callout-warn">
That difference is exactly what makes naive regression biased. Any estimator
that omits `W` will absorb part of its association with `Y` into the treatment
coefficient.
</div>

---

## Step 3: Define Estimators

An estimator is any function that accepts the data frame returned by
`causalsim_draw()` and returns a named numeric vector. The `ci_lower` and
`ci_upper` fields are optional but enable coverage and power metrics.

```{r estimators}
# Naive: regresses Y on A only, omits the confounder
naive_est <- function(data) {
  fit <- lm(Y ~ A, data = data)
  est <- coef(fit)[["A"]]
  se <- sqrt(vcov(fit)["A", "A"])
  c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
}

# OLS: adjusts for the observed confounder W
ols_est <- function(data) {
  fit <- lm(Y ~ A + W, data = data)
  est <- coef(fit)[["A"]]
  se <- sqrt(vcov(fit)["A", "A"])
  c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
}
```

Named lists are also accepted, so the following is equivalent:

```{r estimator-list, eval = FALSE}
ols_est <- function(data) {
  fit <- lm(Y ~ A + W, data = data)
  ci <- confint(fit)["A", ]
  list(
    estimate = coef(fit)["A"],
    ci_lower = ci[1],
    ci_upper = ci[2]
  )
}
```

---

## Step 4: Evaluate Each Estimator

`causalsim_eval()` runs `reps` independent replications and returns a tidy
summary of bias, RMSE, coverage, and power with Monte Carlo standard errors.

```{r eval-naive}
eval_naive <- causalsim_eval(dgp, naive_est, reps = 300L, seed = 1L)
eval_naive
```

```{r eval-ols}
eval_ols <- causalsim_eval(dgp, ols_est, reps = 300L, seed = 1L)
eval_ols
```

The naive estimator's bias is substantial: treatment is positively correlated
with $W$, which also raises $Y$ through the baseline, so the unadjusted
coefficient absorbs part of that association. OLS eliminates the bias by
conditioning on $W$. Coverage for the naive estimator falls well below the
nominal 95% because the confidence intervals are centered on the wrong value.

`summary()` adds the full distribution of per-replication estimates, and
`plot()` shows it as a histogram:

```{r summary-ols}
summary(eval_ols)
```

```{r plot-ols, fig.cap = "Distribution of OLS estimates over 300 replications. Solid line: true ATE. Dashed line: mean estimate."}
plot(eval_ols)
```

---

## Step 5: Vary Sample Size with `causalsim_eval_grid()`

`causalsim_eval_grid()` evaluates an estimator over the Cartesian product of the
supplied parameter values, returning a tidy data frame of metrics for each cell.
Here we vary `n` across four levels to track how the OLS estimator's precision
improves with more data.

```{r grid-n}
grid_n <- causalsim_eval_grid(
  dgp = dgp,
  estimator = ols_est,
  vary = list(n = c(100L, 250L, 500L, 1000L)),
  reps = 300L,
  metrics = c("bias", "rmse"),
  seed = 1L
)
grid_n
```

RMSE roughly halves as $n$ quadruples, consistent with $\sqrt{n}$-rate
convergence for OLS in a correctly specified model. Bias stays near zero at
every sample size.

---

## Step 6: Vary Confounding Strength

Varying the `propensity` preset shows how bias scales with confounding. Because
`causalsim_eval_grid()` accepts one estimator at a time, we run it separately and
combine the results.

```{r grid-confounding}
conf_levels <- list(propensity = c("low", "moderate", "high"))

grid_naive <- causalsim_eval_grid(dgp, naive_est,
                             vary = conf_levels,
                             reps = 300L,
                             metrics = "bias",
                             seed = 1L)

grid_ols <- causalsim_eval_grid(dgp, ols_est,
                           vary = conf_levels,
                           reps = 300L,
                           metrics = "bias",
                           seed = 1L)

comparison <- rbind(
  cbind(estimator = "naive", grid_naive$results),
  cbind(estimator = "ols",   grid_ols$results)
)
comparison <- comparison[order(comparison$propensity, comparison$estimator), ]
rownames(comparison) <- NULL
comparison
```

Naive bias grows proportionally with confounding strength. OLS remains near zero
across all three levels because $W$ is observed and included in the model.

---

## Step 7: Heterogeneous Effects (Effect Modifiers)

So far the treatment effect has been constant. Real effects often vary across
subgroups. Declare a covariate with `role = "effect_modifier"` and reference it
in a function passed to `effect`:

```{r het-dgp}
het_dgp <- causalsim_dgp(
  n = 4000,
  covariates = list(
    W = causalsim_covar("normal", role = "confounder"),
    V = causalsim_covar("binary", role = "effect_modifier", prob = 0.5)
  ),
  effect = function(V) 2 + 3 * V,   # effect is 2 when V = 0, 5 when V = 1
  propensity = function(W) plogis(0.5 * W),
  baseline = function(W) W
)
het_dgp
```

The function passed to `effect` is what activates the modifier. A covariate
labelled `effect_modifier` but never referenced by `effect` is inert, and
`causalsim_dgp()` warns when that happens — so the role can never silently do
nothing.

Ground truth is carried per unit in `.tau`, so the subgroup effects are known
exactly:

```{r het-truth}
d <- causalsim_draw(het_dgp, seed = 1L)
tapply(d$.tau, d$V, mean)   # 2 for V = 0, 5 for V = 1
```

An estimator that ignores the modifier recovers only the overall average
effect, while one that interacts treatment with `V` recovers the subgroup
effects:

```{r het-estimators}
overall <- lm(Y ~ A + W, data = d)         # assumes a constant effect
interact <- lm(Y ~ A * V + W, data = d)     # allows the effect to vary with V

c(
  average = coef(overall)[["A"]],
  subgroup_v0 = coef(interact)[["A"]],
  subgroup_v1 = coef(interact)[["A"]] + coef(interact)[["A:V"]]
)
```

The constant-effect model lands near the true ATE (`r round(het_dgp$true_ate, 2)`),
but hides the heterogeneity; the interaction model recovers both subgroup
effects. This is the setup for benchmarking CATE / heterogeneous-effect
estimators (causal forests, meta-learners): because the true subgroup effects
are known, any estimator's recovery of them can be scored.

---

## Where to go next

This workflow (define, evaluate, grid) scales to more complex settings. A few
directions:

| Goal | How |
|---|---|
| Just generate data | Use `causalsim()` for a single dataset in one call |
| Heterogeneous effects | See Step 7 — declare an `effect_modifier` and pass a function to `effect` |
| Non-normal covariates | Use `causalsim_covar("binary")` or `causalsim_covar("uniform")` in `covariates` |
| Multiple confounders | Set `n_confounders = 3` or pass named `covariates` |
| Custom covariate structure | Mix `n_confounders` with explicit `covariates = list(...)` |

See `?causalsim`, `?causalsim_dgp`, and `?causalsim_covar` for the full API.

---

## Session info

```{r session-info}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causalsim.R
\name{causalsim}
\alias{causalsim}
\title{Simulate a causal dataset with known ground truth}
\usage{
causalsim(
  n,
  effect = 1,
  propensity = "moderate",
  baseline = "moderate",
  sigma = 1,
  covariates = list(),
  n_confounders = 0L,
  n_effect_modifiers = 0L,
  n_noise = 0L,
  mc_draws = 10000L,
  seed = NULL
)
}
\arguments{
\item{n}{Positive integer. Sample size for each simulated dataset.}

\item{effect}{Numeric scalar or function. A scalar specifies a constant
(homogeneous) treatment effect; ATE = CATE everywhere. A function should
accept named arguments matching covariate names defined in the DGP and
return a numeric vector of individual-level causal effects (CATE).
See Details.}

\item{propensity}{Numeric scalar, preset string, or function. Treatment
assignment probability. A scalar (e.g. \code{0.5}) gives a constant propensity
(randomized trial). Preset strings \code{"low"}, \code{"moderate"}, \code{"high"} generate
a logistic propensity over confounders with coefficients 0.25, 0.5, and
1.0 respectively. A function follows the same named-argument convention as
\code{effect} and must return values in [0, 1]. Defaults to \code{"moderate"}.}

\item{baseline}{Numeric scalar, preset string, or function. Mean potential
outcome under control, \code{E[Y(0) | W]}. Preset strings follow the same
levels as \code{propensity} and apply a linear combination of confounders.
Defaults to \code{"moderate"}, so that confounders declared via \code{n_confounders}
(or \code{role = "confounder"}) enter both the treatment and outcome models and
therefore actually induce confounding bias. With no confounders present,
any preset baseline resolves to \code{0}. Set a numeric scalar (e.g. \code{0}) for a
constant baseline that ignores covariates.}

\item{sigma}{Positive numeric. Standard deviation of the outcome noise term.
Default \code{1}.}

\item{covariates}{Named list of \code{\link[=causalsim_covar]{causalsim_covar()}} objects (Option A /
explicit path).
Each name becomes the column name in generated data and the argument name
expected by \code{effect}, \code{propensity}, and \code{baseline} functions. Merged with
any auto-generated covariates; name collisions error.}

\item{n_confounders}{Non-negative integer. Standard normal confounders
auto-generated as \code{W} (single) or \verb{W1, W2, ...} (multiple).}

\item{n_effect_modifiers}{Non-negative integer. Auto-generates standard
normal effect modifiers as \code{V} or \verb{V1, V2, ...}.}

\item{n_noise}{Non-negative integer. Auto-generates standard normal noise
covariates as \code{X} or \verb{X1, X2, ...}.}

\item{mc_draws}{Positive integer. Monte Carlo draws for true ATE
approximation. Default \code{10000L}. Ignored for scalar \code{effect}.}

\item{seed}{Integer or \code{NULL}. Passed to \code{\link[=set.seed]{set.seed()}} before drawing.
Default \code{NULL}.}
}
\value{
A data frame with covariate columns, \code{A}, \code{Y}, \code{.tau}, and \code{.p}.
}
\description{
A convenience wrapper around \code{\link[=causalsim_dgp]{causalsim_dgp()}} and \code{\link[=causalsim_draw]{causalsim_draw()}}.
Defines a causal data generating process and returns one simulated dataset
in a single call.
}
\details{
The returned data frame includes all covariate columns, the treatment
indicator \code{A}, the outcome \code{Y}, and two ground-truth columns:

\describe{
\item{\code{.tau}}{Individual-level causal effect (CATE).}
\item{\code{.p}}{True propensity score.}
}

For full control over the DGP (multiple draws, grid evaluation, estimator
benchmarking), use \code{\link[=causalsim_dgp]{causalsim_dgp()}} and \code{\link[=causalsim_draw]{causalsim_draw()}} directly.
}
\examples{
# One confounder, constant effect of 2, moderate confounding
data <- causalsim(n = 500, n_confounders = 1, effect = 2, seed = 1L)
head(data)

# Heterogeneous effect with an explicit covariate spec
data2 <- causalsim(
  n = 500,
  covariates = list(
    W = causalsim_covar("normal", role = "confounder"),
    V = causalsim_covar("binary", role = "effect_modifier", prob = 0.4)
  ),
  effect = function(V) 2 + 1.5 * V,
  propensity = function(W) plogis(0.5 * W),
  seed = 42L
)
head(data2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causalsim-package.R
\docType{package}
\name{causalsim-package}
\alias{causalsim-package}
\title{causalsim: Simulation-Ready Causal Data Generating Processes}
\description{
Construct, simulate, and evaluate causal data generating processes (DGPs)
with known ground truth. Covariate roles (confounder, effect modifier,
noise) and heterogeneous treatment effects are first-class concepts, and
estimator performance is summarised with bias, RMSE, confidence-interval
coverage, and power. Results are returned as plain data frames for further
analysis or plotting.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://chaycereed.github.io/causalsim/}
  \item \url{https://github.com/chaycereed/causalsim}
  \item Report bugs at \url{https://github.com/chaycereed/causalsim/issues}
}

}
\author{
\strong{Maintainer}: Chayce Reed \email{Chayce.Reed.HSE@dartmouth.edu}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causalsim_covar.R
\name{causalsim_covar}
\alias{causalsim_covar}
\title{Define a covariate for a causal DGP}
\usage{
causalsim_covar(dist = "normal", role = "confounder", ...)
}
\arguments{
\item{dist}{Character. Distribution family. One of \code{"normal"}, \code{"binary"},
\code{"uniform"}.}

\item{role}{Character vector. Causal role(s). One or more of:
\describe{
\item{\code{"confounder"}}{Enters both the propensity model and the outcome
baseline; creates confounding bias in naive estimators.}
\item{\code{"effect_modifier"}}{Available for use in the \code{effect} function;
excluded from the propensity model unless also a \code{"confounder"}.}
\item{\code{"noise"}}{Independent of both treatment and outcome; adds
variance without confounding.}
}
Roles are not mutually exclusive: \code{role = c("confounder", "effect_modifier")}
specifies a variable that both confounds and moderates the effect.}

\item{...}{Distribution parameters.
\itemize{
\item \code{"normal"}: \code{mean} (default 0), \code{sd} (default 1)
\item \code{"binary"}: \code{prob} (default 0.5)
\item \code{"uniform"}: \code{min} (default 0), \code{max} (default 1)
}}
}
\value{
An S3 object of class \code{causalsim_covar}.
}
\description{
Constructs a fully-specified covariate descriptor for use in
\code{\link[=causalsim_dgp]{causalsim_dgp()}}. Every covariate has a distribution family, optional
distribution parameters, and one or more causal roles that control how it
enters the automatically-generated treatment and outcome models.
}
\examples{
# Standard normal confounder
causalsim_covar("normal", role = "confounder", mean = 0, sd = 1)

# Binary effect modifier
causalsim_covar("binary", role = "effect_modifier", prob = 0.4)

# Variable that both confounds and moderates the effect
causalsim_covar("normal", role = c("confounder", "effect_modifier"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causalsim_eval.R
\name{summary.causalsim_eval}
\alias{summary.causalsim_eval}
\title{Summarise a causalsim_eval result}
\usage{
\method{summary}{causalsim_eval}(object, ...)
}
\arguments{
\item{object}{A \code{causalsim_eval} object.}

\item{...}{Ignored.}
}
\value{
An S3 object of class \code{causalsim_eval_summary} with components:
\describe{
\item{\code{metrics}}{The metrics data frame from the original result.}
\item{\code{mean_estimate}}{Mean of per-replication estimates.}
\item{\code{sd_estimate}}{Standard deviation of per-replication estimates.}
\item{\code{median_estimate}}{Median of per-replication estimates.}
\item{\code{p10_estimate}}{10th percentile of per-replication estimates.}
\item{\code{p90_estimate}}{90th percentile of per-replication estimates.}
\item{\code{true_ate}}{True ATE from the DGP.}
\item{\code{reps}}{Number of replications.}
}
}
\description{
Returns a structured summary of estimator performance including the metrics
table and the distribution of per-replication estimates (mean, SD, median,
and 10th / 90th percentiles).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causalsim_eval.R
\name{plot.causalsim_eval}
\alias{plot.causalsim_eval}
\title{Plot the distribution of estimates from a causalsim_eval result}
\usage{
\method{plot}{causalsim_eval}(x, ...)
}
\arguments{
\item{x}{A \code{causalsim_eval} object.}

\item{...}{Additional arguments passed to \code{\link[graphics:hist]{graphics::hist()}}.}
}
\value{
\code{x}, invisibly.
}
\description{
Draws a histogram of per-replication estimates with vertical lines marking
the true ATE (solid) and the mean estimate (dashed).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causalsim_dgp.R
\name{causalsim_dgp}
\alias{causalsim_dgp}
\title{Create a causal data generating process}
\usage{
causalsim_dgp(
  n,
  effect = 1,
  propensity = "moderate",
  baseline = "moderate",
  sigma = 1,
  covariates = list(),
  n_confounders = 0L,
  n_effect_modifiers = 0L,
  n_noise = 0L,
  mc_draws = 10000L
)
}
\arguments{
\item{n}{Positive integer. Sample size for each simulated dataset.}

\item{effect}{Numeric scalar or function. A scalar specifies a constant
(homogeneous) treatment effect; ATE = CATE everywhere. A function should
accept named arguments matching covariate names defined in the DGP and
return a numeric vector of individual-level causal effects (CATE).
See Details.}

\item{propensity}{Numeric scalar, preset string, or function. Treatment
assignment probability. A scalar (e.g. \code{0.5}) gives a constant propensity
(randomized trial). Preset strings \code{"low"}, \code{"moderate"}, \code{"high"} generate
a logistic propensity over confounders with coefficients 0.25, 0.5, and
1.0 respectively. A function follows the same named-argument convention as
\code{effect} and must return values in [0, 1]. Defaults to \code{"moderate"}.}

\item{baseline}{Numeric scalar, preset string, or function. Mean potential
outcome under control, \code{E[Y(0) | W]}. Preset strings follow the same
levels as \code{propensity} and apply a linear combination of confounders.
Defaults to \code{"moderate"}, so that confounders declared via \code{n_confounders}
(or \code{role = "confounder"}) enter both the treatment and outcome models and
therefore actually induce confounding bias. With no confounders present,
any preset baseline resolves to \code{0}. Set a numeric scalar (e.g. \code{0}) for a
constant baseline that ignores covariates.}

\item{sigma}{Positive numeric. Standard deviation of the outcome noise term.
Default \code{1}.}

\item{covariates}{Named list of \code{\link[=causalsim_covar]{causalsim_covar()}} objects (Option A /
explicit path).
Each name becomes the column name in generated data and the argument name
expected by \code{effect}, \code{propensity}, and \code{baseline} functions. Merged with
any auto-generated covariates; name collisions error.}

\item{n_confounders}{Non-negative integer. Standard normal confounders
auto-generated as \code{W} (single) or \verb{W1, W2, ...} (multiple).}

\item{n_effect_modifiers}{Non-negative integer. Auto-generates standard
normal effect modifiers as \code{V} or \verb{V1, V2, ...}.}

\item{n_noise}{Non-negative integer. Auto-generates standard normal noise
covariates as \code{X} or \verb{X1, X2, ...}.}

\item{mc_draws}{Positive integer. Monte Carlo draws for true ATE
approximation. Default \code{10000L}. Ignored for scalar \code{effect}.}
}
\value{
An S3 object of class \code{causalsim_dgp} with components:
\describe{
\item{\code{n}}{Sample size (integer)}
\item{\code{covar_spec}}{Named list of \code{\link[=causalsim_covar]{causalsim_covar()}} objects}
\item{\code{effect_fn}}{Normalized effect function}
\item{\code{propensity_fn}}{Normalized propensity function}
\item{\code{baseline_fn}}{Normalized baseline function}
\item{\code{sigma}}{Outcome noise standard deviation}
\item{\code{true_ate}}{True ATE: exact for scalar, MC approximation otherwise}
\item{\code{heterogeneous}}{Logical; \code{TRUE} if \code{effect} was a function}
\item{\code{mc_draws}}{Monte Carlo draws used (integer)}
}
}
\description{
Defines a causal DGP with known ground truth. Covariates can be specified
via shorthand count arguments (Option B), an explicit named list of
\code{\link[=causalsim_covar]{causalsim_covar()}} objects (Option A), or both combined.
}
\details{
\subsection{Structural model}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{W ~ covariate_spec
A ~ Bernoulli(propensity(W))
Y  = baseline(W) + effect(W) * A + N(0, sigma^2)
}\if{html}{\out{</div>}}

\code{A} is used for treatment throughout to avoid collision with R's built-in
\code{T} alias.
}

\subsection{Function calling convention}{

The \code{effect}, \code{propensity}, and \code{baseline} functions are called with named
arguments matching covariate names, not a data frame. Write:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{effect = function(W) 2 + 1.5 * W
propensity = function(W1, W2) plogis(0.3 * W1 + 0.5 * W2)
}\if{html}{\out{</div>}}

Every argument name is validated against the DGP's covariate spec at
construction time, so mismatches surface immediately rather than at draw
time. The propensity function is additionally evaluated on a small test draw
to confirm it returns values in [0, 1].
}

\subsection{True ATE}{

For scalar \code{effect} the true ATE is exact. For function \code{effect} it is
approximated via Monte Carlo over \code{mc_draws} draws from the covariate
distribution.
}
}
\examples{
# Minimal: one confounder, constant effect, moderate confounding
dgp <- causalsim_dgp(n = 500, n_confounders = 1, effect = 2)
dgp

# Heterogeneous effect, explicit covariate spec
dgp2 <- causalsim_dgp(
  n = 500,
  covariates = list(
    W = causalsim_covar("normal", role = "confounder"),
    V = causalsim_covar("binary", role = "effect_modifier", prob = 0.4)
  ),
  effect = function(V) 2 + 1.5 * V,
  propensity = function(W) plogis(0.5 * W),
  baseline = function(W) 1.5 * W
)

# Mixed: shorthand confounders + explicit noise covariate + RCT propensity
dgp3 <- causalsim_dgp(
  n = 1000,
  n_confounders = 2,
  covariates = list(X = causalsim_covar("normal", role = "noise")),
  effect = 1,
  propensity = 0.5
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causalsim_eval.R
\name{causalsim_eval}
\alias{causalsim_eval}
\title{Evaluate a causal estimator against a known DGP}
\usage{
causalsim_eval(
  dgp,
  estimator,
  reps = 200L,
  metrics = c("bias", "rmse", "coverage", "power"),
  seed = NULL
)
}
\arguments{
\item{dgp}{A \code{causalsim_dgp} object.}

\item{estimator}{A function that accepts a data frame produced by
\code{\link[=causalsim_draw]{causalsim_draw()}} and returns a named numeric vector or single-row
data frame. Must include an \code{estimate} field. For \code{"coverage"} and
\code{"power"} metrics, must also include \code{ci_lower} and \code{ci_upper}. See
Details.}

\item{reps}{Positive integer. Number of simulation replications.
Default \code{200L}.}

\item{metrics}{Character vector. Subset of \code{"bias"}, \code{"rmse"},
\code{"coverage"}, \code{"power"}. Default: all four.}

\item{seed}{Integer or \code{NULL}. Passed to \code{\link[=set.seed]{set.seed()}} before the first
replication. Default \code{NULL}.}
}
\value{
An S3 object of class \code{causalsim_eval} with components:
\describe{
\item{\code{metrics}}{Tidy data frame: \code{metric}, \code{value}, \code{se}.}
\item{\code{draws}}{Data frame of per-replication estimates, one row per rep.}
\item{\code{true_ate}}{The DGP's true ATE.}
\item{\code{reps}}{Number of replications run.}
}
}
\description{
Repeatedly draws datasets from a \code{\link[=causalsim_dgp]{causalsim_dgp()}} object, applies a
user-supplied estimator to each draw, and returns a tidy summary of
estimator performance against the known ground truth.
}
\details{
\subsection{Estimator convention}{

The estimator receives the full data frame returned by \code{\link[=causalsim_draw]{causalsim_draw()}},
including ground-truth columns \code{.tau} and \code{.p}. It should return a named
numeric vector or one-row data frame with at minimum:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{my_estimator <- function(data) \{
  fit <- lm(Y ~ A + W, data = data)
  est <- coef(fit)["A"]
  se <- sqrt(vcov(fit)["A", "A"])
  c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
\}
}\if{html}{\out{</div>}}
}

\subsection{Metrics}{

\describe{
\item{\code{"bias"}}{Monte Carlo estimate of bias: \code{mean(estimate) - true_ate}.}
\item{\code{"rmse"}}{Root mean squared error: \code{sqrt(mean((estimate - true_ate)^2))}.}
\item{\code{"coverage"}}{Proportion of replications where the CI brackets
\code{true_ate}. Requires \code{ci_lower} and \code{ci_upper}.}
\item{\code{"power"}}{Proportion of replications where the CI excludes zero,
i.e., the rejection rate of H0: ATE = 0. When \code{true_ate != 0},
this is power; when \code{true_ate == 0}, it is the Type I error rate.
Requires \code{ci_lower} and \code{ci_upper}.}
}

Monte Carlo standard errors (MCSE) are reported alongside each metric.
For bias: \code{sd(estimates) / sqrt(reps)}. For RMSE: delta-method
approximation. For coverage and power: Bernoulli SE.
}

\subsection{Raw draws}{

Per-replication estimates are stored in \code{result$draws} and can be used
for custom analysis or plotting.
}
}
\examples{
dgp <- causalsim_dgp(n = 300, n_confounders = 1, effect = 2,
                     propensity = 0.5)

# OLS estimator (unbiased under RCT)
ols_estimator <- function(data) {
  fit <- lm(Y ~ A + W, data = data)
  est <- coef(fit)[["A"]]
  se <- sqrt(vcov(fit)["A", "A"])
  c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
}

\donttest{
result <- causalsim_eval(dgp, ols_estimator, reps = 100L, seed = 1L)
result
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causalsim_draw.R
\name{causalsim_draw}
\alias{causalsim_draw}
\title{Draw a dataset from a causal DGP}
\usage{
causalsim_draw(dgp, seed = NULL)
}
\arguments{
\item{dgp}{A \code{causalsim_dgp} object created by \code{\link[=causalsim_dgp]{causalsim_dgp()}}.}

\item{seed}{Integer or \code{NULL}. If non-null, passed to \code{\link[=set.seed]{set.seed()}} before
generating any random values, making the draw reproducible. Default \code{NULL}.}
}
\value{
A data frame with \code{dgp$n} rows and the following columns:
\describe{
\item{Covariate columns}{One column per covariate in \code{dgp$covar_spec},
named to match the spec (e.g. \code{W}, \code{W1}, \code{Z}).}
\item{\code{A}}{Binary treatment indicator (0/1).}
\item{\code{Y}}{Observed outcome.}
\item{\code{.tau}}{Individual treatment effect (CATE). Ground truth.}
\item{\code{.p}}{Individual propensity score. Ground truth.}
}

The \code{.tau} and \code{.p} columns carry individual-level ground truth and are
prefixed with \code{.} to distinguish them from observed variables.
\code{causalsim_eval()} uses them directly without re-estimation.
}
\description{
Simulates one dataset from a \code{\link[=causalsim_dgp]{causalsim_dgp()}} object using the structural
model:
}
\details{
\if{html}{\out{<div class="sourceCode">}}\preformatted{W ~ covariate_spec
A ~ Bernoulli(propensity(W))
Y  = baseline(W) + effect(W) * A + N(0, sigma^2)
}\if{html}{\out{</div>}}
}
\examples{
dgp <- causalsim_dgp(n = 500, n_confounders = 1, effect = 2)
d   <- causalsim_draw(dgp, seed = 1)
head(d)

# Reproducible draws
d1 <- causalsim_draw(dgp, seed = 42)
d2 <- causalsim_draw(dgp, seed = 42)
identical(d1, d2)  # TRUE

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causalsim_eval_grid.R
\name{causalsim_eval_grid}
\alias{causalsim_eval_grid}
\title{Evaluate an estimator across a parameter grid}
\usage{
causalsim_eval_grid(
  dgp,
  estimator,
  vary,
  reps = 200L,
  metrics = c("bias", "rmse", "coverage", "power"),
  seed = NULL,
  verbose = FALSE
)
}
\arguments{
\item{dgp}{A \code{causalsim_dgp} object. Provides fixed parameter values for
all dimensions not listed in \code{vary}.}

\item{estimator}{A function as accepted by \code{\link[=causalsim_eval]{causalsim_eval()}}.}

\item{vary}{Named list of atomic vectors. Each name must be a valid
\code{\link[=causalsim_dgp]{causalsim_dgp()}} argument (except \code{covariates}, which cannot be
varied atomically). Each vector supplies the values to try for that
dimension. The full grid is the Cartesian product of all dimensions.}

\item{reps}{Positive integer. Replications per grid cell. Default \code{200L}.}

\item{metrics}{Character vector. Passed to \code{\link[=causalsim_eval]{causalsim_eval()}}.}

\item{seed}{Integer or \code{NULL}. Passed to \code{\link[=set.seed]{set.seed()}} before the first
cell. The same RNG stream continues across cells, so results are
jointly reproducible. Default \code{NULL}.}

\item{verbose}{Logical. If \code{TRUE}, prints a progress message before each
cell. Default \code{FALSE}.}
}
\value{
An S3 object of class \code{causalsim_eval_grid} with components:
\describe{
\item{\code{results}}{Tidy data frame: one row per (cell, metric).
Columns: grid parameter names, \code{metric}, \code{value}, \code{se}.}
\item{\code{grid}}{Data frame of grid cells, one row per cell.}
\item{\code{vary}}{Character vector of varied parameter names.}
\item{\code{reps}}{Replications per cell.}
\item{\code{metrics}}{Metrics evaluated.}
}
}
\description{
Runs \code{\link[=causalsim_eval]{causalsim_eval()}} over the Cartesian product of one or more DGP
parameter values, returning a tidy data frame of performance metrics for
every cell. Designed for studying how estimator behavior changes with
sample size, confounding strength, effect size, or noise level.
}
\details{
\subsection{How it works}{

For each cell in \code{expand.grid(vary)}, \code{causalsim_eval_grid()} takes \code{dgp}'s
stored original parameters, overrides the cell's values, reconstructs a
new \code{\link[=causalsim_dgp]{causalsim_dgp()}}, runs \code{\link[=causalsim_eval]{causalsim_eval()}}, and tags the resulting
metrics with the cell's parameter values.
}

\subsection{Constraints on \code{vary}}{

Elements of \code{vary} must be atomic vectors (character, numeric, integer).
Functions cannot be varied via this interface. \code{covariates} (a list of
\code{\link[=causalsim_covar]{causalsim_covar()}} objects) is excluded. For complex covariate variation,
construct DGPs manually and use \code{\link[=causalsim_eval]{causalsim_eval()}} directly.

Note that varying \code{n_confounders} from 1 to 2 changes auto-generated
covariate names from \code{W} to \verb{W1, W2}. If the base DGP's \code{effect} function
references \code{W}, the reconstructed DGP will error at construction time.
Design the base DGP accordingly.
}
}
\examples{
dgp <- causalsim_dgp(n = 500, n_confounders = 1, effect = 2,
                     propensity = 0.5)

ols_estimator <- function(data) {
  fit <- lm(Y ~ A + W, data = data)
  est <- coef(fit)[["A"]]
  se  <- sqrt(vcov(fit)["A", "A"])
  c(estimate = est, ci_lower = est - 1.96 * se, ci_upper = est + 1.96 * se)
}

\donttest{
grid_result <- causalsim_eval_grid(
  dgp = dgp,
  estimator = ols_estimator,
  vary = list(n = c(250L, 500L, 1000L)),
  reps = 50L,
  seed = 1L
)
grid_result
}

}
